from __future__ import annotations

from abc import ABC
from functools import reduce
from typing import Optional, List, TYPE_CHECKING, Set, Tuple, Union, TypeVar, Type, overload

import shapely.geometry as shp
import shapely.geometry.base as shpb
from pymeos_cffi import *

from .tbool import TBool
from .tfloat import TFloat, TFloatInst, TFloatSeq, TFloatSeqSet
from ..temporal import Temporal, TInstant, TSequence, TSequenceSet, TInterpolation
from ..collections import *

if TYPE_CHECKING:
    from ..boxes import STBox, Box
    from geopandas import GeoDataFrame


def import_geopandas():
    try:
        import geopandas as gpd
        return gpd
    except ImportError:
        print('Geopandas not found. Please install geopandas to use this function.')
        raise


TG = TypeVar('TG', bound='TPoint')
TI = TypeVar('TI', bound='TPointInst')
TS = TypeVar('TS', bound='TPointSeq')
TSS = TypeVar('TSS', bound='TPointSeqSet')
Self = TypeVar('Self', bound='TPoint')
TF = TypeVar('TF', bound='TFloat', covariant=True)


class TPoint(Temporal[shp.Point, TG, TI, TS, TSS], ABC):
    """
    Abstract class for temporal points.
    """

    # ------------------------- Constructors ----------------------------------
    def __init__(self, _inner) -> None:
        super().__init__()

    @classmethod
    def from_hexwkb(cls: Type[Self], hexwkb: str, srid: Optional[int] = None) -> Self:
        result = super().from_hexwkb(hexwkb)
        return result.set_srid(srid) if srid is not None else result

    # ------------------------- Output ----------------------------------------
    def __str__(self):
        """
        Returns the string representation of the temporal point.

        Returns:
            A new :class:`str` representing the temporal point.

        MEOS Functions:
            tpoint_out
        """
        return tpoint_as_text(self._inner, 15)

    def as_wkt(self, precision: int = 15) -> str:
        """
        Returns the temporal point as a WKT string.

        Args:
            precision: The precision of the returned geometry.

        Returns:
            A new :class:`str` representing the temporal point.

        MEOS Functions:
            tpoint_out
        """
        return tpoint_as_text(self._inner, precision)

    def as_ewkt(self, precision: int = 15) -> str:
        """
        Returns the temporal point as an EWKT string.

        Args:
            precision: The precision of the returned geometry.

        Returns:
            A new :class:`str` representing the temporal point  .

        MEOS Functions:
            tpoint_as_ewkt
        """
        return tpoint_as_ewkt(self._inner, precision)

    def as_geojson(self, option: int = 1, precision: int = 15, srs: Optional[str] = None) -> str:
        """
        Returns the trajectory of the temporal point as a GeoJSON string.

        Args:
            option: The option to use when serializing the trajectory.
            precision: The precision of the returned geometry.
            srs: The spatial reference system of the returned geometry.

        Returns:
            A new GeoJSON string representing the trajectory of the temporal point.

        MEOS Functions:
            gserialized_as_geojson
        """
        return gserialized_as_geojson(tpoint_trajectory(self._inner), option, precision, srs)

    def to_shapely_geometry(self, precision: int = 15) -> shpb.BaseGeometry:
        """
        Returns the trajectory of the temporal point as a Shapely geometry.

        Args:
            precision: The precision of the returned geometry.

        Returns:
            A new :class:`~shapely.geometry.base.BaseGeometry` representing the
            trajectory.

        MEOS Functions:
            gserialized_to_shapely_geometry
        """
        return gserialized_to_shapely_geometry(tpoint_trajectory(self._inner),
                                               precision)

    # ------------------------- Accessors -------------------------------------
    def bounding_box(self) -> STBox:
        """
        Returns the bounding box of the `self`.

        Returns:
            An :class:`~pymeos.boxes.STBox` representing the bounding box.

        MEOS Functions:
            tpoint_to_stbox
        """
        from ..boxes import STBox
        return STBox(_inner=tpoint_to_stbox(self._inner))

    def values(self, precision: int = 15) -> List[shp.Point]:
        """
        Returns the values of the temporal point.

        Returns:
            A :class:`list` of :class:`~shapely.geometry.Point` with the values.

        MEOS Functions:
            tpoint_values
        """
        result, count = tpoint_values(self._inner)
        return [gserialized_to_shapely_point(result[i], precision) for i in range(count)]

    def start_value(self, precision: int = 15) -> shp.Point:
        """
        Returns the start value of the temporal point.

        Returns:
            A :class:`~shapely.geometry.Point` with the start value.

        MEOS Functions:
            tpoint_start_value
        """
        return gserialized_to_shapely_point(tpoint_start_value(self._inner), precision)

    def end_value(self, precision: int = 15) -> shp.Point:
        """
        Returns the end value of the temporal point.

        Returns:
            A :class:`~shapely.geometry.Point` with the end value.

        MEOS Functions:
            tpoint_end_value
        """
        return gserialized_to_shapely_point(tpoint_end_value(self._inner), precision)

    def value_set(self, precision: int = 15) -> Set[shp.Point]:
        """
        Returns the set of values of `self`.
        Note that when the interpolation is linear, the set will contain only the waypoints.

        Returns:
            A :class:`set` of :class:`~shapely.geometry.Point` with the values.

        MEOS Functions:
            tpoint_values
        """
        values, count = tpoint_values(self._inner)
        return {gserialized_to_shapely_point(values[i], precision) for i in range(count)}

    def value_at_timestamp(self, timestamp: datetime, precision: int = 15) -> shp.Point:
        """
        Returns the value of the temporal point at the given timestamp.

        Args:
            timestamp: A :class:`datetime` representing the timestamp.
            precision: An :class:`int` representing the precision of the coordinates.

        Returns:
            A :class:`~shapely.geometry.Point` with the value.

        MEOS Functions:
            tpoint_value_at_timestamp
        """
        return gserialized_to_shapely_point(
            tpoint_value_at_timestamp(self._inner, datetime_to_timestamptz(timestamp), True)[0], precision)

    def length(self) -> float:
        """
        Returns the length of the trajectory.

        Returns:
            A :class:`float` with the length of the trajectory.

        MEOS Functions:
            tpoint_length
        """
        return tpoint_length(self._inner)

    def cumulative_length(self) -> TFloat:
        """
        Returns the cumulative length of the trajectory.

        Returns:
            A :class:`TFloat` with the cumulative length of the trajectory.

        MEOS Functions:
            tpoint_cumulative_length
        """
        result = tpoint_cumulative_length(self._inner)
        return Temporal._factory(result)

    def speed(self) -> TFloat:
        """
        Returns the speed of the temporal point.

        Returns:
            A :class:`TFloat` with the speed of the temporal point.

        MEOS Functions:
            tpoint_speed
        """
        result = tpoint_speed(self._inner)
        return Temporal._factory(result)

    def x(self) -> TF:
        """
        Returns the x coordinate of the temporal point.

        Returns:
            A :class:`TFloat` with the x coordinate of the temporal point.

        MEOS Functions:
            tpoint_get_x
        """
        result = tpoint_get_x(self._inner)
        return Temporal._factory(result)

    def y(self) -> TF:
        """
        Returns the y coordinate of the temporal point.

        Returns:
            A :class:`TFloat` with the y coordinate of the temporal point.

        MEOS Functions:
            tpoint_get_y
        """
        result = tpoint_get_y(self._inner)
        return Temporal._factory(result)

    def z(self) -> TF:
        """
        Returns the z coordinate of the temporal point.

        Returns:
            A :class:`TFloat` with the z coordinate of the temporal point.

        MEOS Functions:
            tpoint_get_z
        """
        result = tpoint_get_z(self._inner)
        return Temporal._factory(result)

    def has_z(self) -> bool:
        """
        Returns whether the temporal point has a z coordinate.

        Returns:
            A :class:`bool` indicating whether the temporal point has a z coordinate.

        MEOS Functions:
            tpoint_start_value
        """
        return self.bounding_box().has_z()

    def stboxes(self) -> List[STBox]:
        """
        Returns a collection of :class:`STBox`es representing the bounding boxes of the segments of the temporal point.

        Returns:
            A :class:`list` of :class:`STBox`es.

        MEOS Functions:
            tpoint_stboxes
        """
        from ..boxes import STBox
        result, count = tpoint_stboxes(self._inner)
        return [STBox(_inner=result + i) for i in range(count)]

    def is_simple(self) -> bool:
        """
        Returns whether the temporal point is simple. That is, whether it does not self-intersect.

        Returns:
            A :class:`bool` indicating whether the temporal point is simple.

        MEOS Functions:
            tpoint_is_simple
        """
        return tpoint_is_simple(self._inner)

    def bearing(self, other: Union[shpb.BaseGeometry, TPoint]) -> TFloat:
        """
        Returns the temporal bearing between the temporal point and `other`.

        Args:
            other: An object to check the bearing to.

        Returns:
            A new :class:`TFloat` indicating the temporal bearing between the temporal point and `other`.

        MEOS Functions:
            bearing_tpoint_point, bearing_tpoint_tpoint
        """
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = bearing_tpoint_point(self._inner, gs, False)
        elif isinstance(other, TPoint):
            result = bearing_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def direction(self) -> float:
        """
        Returns the azimuth of the temporal point between the start and end locations.

        Returns:
            A new :class:`TFloatSeqSet` indicating the direction of the temporal point.

        MEOS Functions:
            tpoint_direction
        """
        return tpoint_direction(self._inner)

    def azimuth(self) -> TFloatSeqSet:
        """
        Returns the temporal azimuth of the temporal point.

        Returns:
            A new :class:`TFloatSeqSet` indicating the temporal azimuth of the temporal point.

        MEOS Functions:
            tpoint_azimuth
        """
        result = tpoint_azimuth(self._inner)
        return Temporal._factory(result)

    def angular_difference(self) -> TFloatSeqSet:
        """
        Returns the angular_difference of the temporal point.

        Returns:
            A new :class:`TFloatSeqSet` indicating the temporal angular_difference of the temporal point.

        MEOS Functions:
            tpoint_angular_difference
        """
        result = tpoint_angular_difference(self._inner)
        return Temporal._factory(result)

    def time_weighted_centroid(self, precision: int = 15) -> shp.Point:
        """
        Returns the time weighted centroid of the temporal point.

        Args:
            precision: The precision of the returned geometry.

        Returns:
            A new :class:`~shapely.geometry.base.BaseGeometry` indicating the time weighted centroid of the temporal point.

        MEOS Functions:
            tpoint_twcentroid
        """
        return gserialized_to_shapely_geometry(tpoint_twcentroid(self._inner), precision)  # type: ignore

    # ------------------------- Spatial Reference System ----------------------
    def srid(self) -> int:
        """
        Returns the SRID.

        Returns:
            An :class:`int` representing the SRID.

        MEOS Functions:
            tpoint_srid
        """
        return tpoint_srid(self._inner)

    def set_srid(self: Self, srid: int) -> Self:
        """
        Returns a new TPoint with the given SRID.


        """
        return self.__class__(_inner=tpoint_set_srid(self._inner, srid))

    # ------------------------- Transformations -------------------------------
    def round(self, maxdd: int = 0) -> TPoint:
        """
        Round the coordinate values to a number of decimal places.

        Returns:
            A new :class:`TGeomPoint` object.

        MEOS Functions:
            tpoint_round
        """
        result = tpoint_round(self._inner, maxdd)
        return Temporal._factory(result)

    def make_simple(self) -> List[TPoint]:
        """
        Split the temporal point into a collection of simple temporal points.
        
        Returns:
            A :class:`list` of :class:`TPoint`es.
            
        MEOS Functions:
            tpoint_make_simple
        """
        result, count = tpoint_make_simple(self._inner)
        from ..factory import _TemporalFactory
        return [_TemporalFactory.create_temporal(result[i]) for i in range(count)]

    def expand(self, other: Union[int, float]) -> STBox:
        """
        Expands ``self`` with `other`.
        The result is equal to ``self`` but with the spatial dimensions
        expanded by `other` in all directions.

        Args:
            other: The object to expand ``self`` with.

        Returns:
            A new :class:`STBox` instance.

        MEOS Functions:
            tpoint_expand_space
        """
        from ..boxes import STBox
        result = tpoint_expand_space(self._inner, float(other))
        return STBox(_inner=result)

    # ------------------------- Restrictions ----------------------------------
    def at(self, other: Union[shpb.BaseGeometry, GeoSet, STBox, Time]) -> TG:
        """
        Returns a new temporal object with the values of `self` restricted to `other`.

        Args:
            other: An object to restrict the values of `self` to.

        Returns:
            A new :TPoint: with the values of `self` restricted to `other`.

        MEOS Functions:
            tpoint_at_value, tpoint_at_stbox, temporal_at_values,
            temporal_at_timestamp, temporal_at_timestampset, temporal_at_period, temporal_at_periodset
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = tpoint_at_value(self._inner, gs)
        elif isinstance(other, GeoSet):
            result = temporal_at_values(self._inner, other._inner)
        elif isinstance(other, STBox):
            result = tpoint_at_stbox(self._inner, other._inner, True)
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[shpb.BaseGeometry, GeoSet, STBox, Time]) -> TG:
        """
        Returns a new temporal object with the values of `self` restricted to the complement of `other`.

        Args:
            other: An object to restrict the values of `self` to the complement of.

        Returns:
            A new :TPoint: with the values of `self` restricted to the complement of `other`.

        MEOS Functions:
            tpoint_minus_value, tpoint_minus_stbox, temporal_minus_values,
            temporal_minus_timestamp, temporal_minus_timestampset, temporal_minus_period, temporal_minus_periodset
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = tpoint_minus_value(self._inner, gs)
        elif isinstance(other, GeoSet):
            result = temporal_minus_values(self._inner, other._inner)
        elif isinstance(other, STBox):
            result = tpoint_minus_stbox(self._inner, other._inner, True)
        else:
            return super().minus(other)
        return Temporal._factory(result)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is left to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if left, False otherwise.

        See Also:
            :meth:`Period.is_before`
        """
        return self.bounding_box().is_left(other)

    def is_over_or_left(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is over or left to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if over or left, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.bounding_box().is_over_or_left(other)

    def is_right(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is right to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if right, False otherwise.

        See Also:
            :meth:`Period.is_after`
        """
        return self.bounding_box().is_right(other)

    def is_over_or_right(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is over or right to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if over or right, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.bounding_box().is_over_or_right(other)

    def is_below(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is below to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if below, False otherwise.

        See Also:
            :meth:`Period.is_before`
        """
        return self.bounding_box().is_below(other)

    def is_over_or_below(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is over or below to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if over or below, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.bounding_box().is_over_or_below(other)

    def is_above(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is above to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if above, False otherwise.

        See Also:
            :meth:`Period.is_after`
        """
        return self.bounding_box().is_above(other)

    def is_over_or_above(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is over or above to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if over or above, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.bounding_box().is_over_or_above(other)

    def is_front(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is front to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if front, False otherwise.

        See Also:
            :meth:`Period.is_before`
        """
        return self.bounding_box().is_front(other)

    def is_over_or_front(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is over or front to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if over or front, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.bounding_box().is_over_or_front(other)

    def is_behind(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is behind to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if behind, False otherwise.

        See Also:
            :meth:`Period.is_after`
        """
        return self.bounding_box().is_behind(other)

    def is_over_or_behind(self, other: Union[Temporal, Box]) -> bool:
        """
        Returns whether the bounding box of `self` is over or behind to the bounding box of `other`.

        Args:
            other: A box or a temporal object to compare to `self`.

        Returns:
            True if over or behind, False otherwise.

        See Also:
            :meth:`Period.is_over_or_before`
        """
        return self.bounding_box().is_over_or_behind(other)

    # ------------------------- Ever Spatial Relationships --------------------
    def is_ever_contained_in(self, container: Union[shpb.BaseGeometry, STBox]) -> bool:
        """
        Returns whether the temporal point is ever contained by `container`.

        Args:
            container: An object to check for containing `self`.

        Returns:
            A :class:`bool` indicating whether the temporal point is ever contained by `container`.

        MEOS Functions:
            econtains_geo_tpoint
        """
        from ..boxes import STBox
        if isinstance(container, shpb.BaseGeometry):
            gs = geo_to_gserialized(container, isinstance(self, TGeogPoint))
            result = econtains_geo_tpoint(gs, self._inner)
        elif isinstance(container, STBox):
            result = econtains_geo_tpoint(stbox_to_geo(container._inner), self._inner)
        else:
            raise TypeError(f'Operation not supported with type {container.__class__}')
        return result == 1

    def is_ever_disjoint(self, other: Union[shpb.BaseGeometry, TPoint, STBox]) -> bool:
        """
        Returns whether the temporal point is ever disjoint from `other`.

        Args:
            other: An object to check for disjointness with.

        Returns:
            A :class:`bool` indicating whether the temporal point is ever disjoint from `other`.

        MEOS Functions:
            edisjoint_tpoint_geo, edisjoint_tpoint_tpoint
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = edisjoint_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            result = edisjoint_tpoint_geo(self._inner, stbox_to_geo(other._inner))
        elif isinstance(other, TPoint):
            result = edisjoint_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return result == 1

    def is_ever_within_distance(self, other: Union[shpb.BaseGeometry, TPoint, STBox],
                                distance: float) -> bool:
        """
        Returns whether the temporal point is ever within `distance` of `other`.

        Args:
            other: An object to check the distance to.
            distance: The distance to check in units of the spatial reference system.

        Returns:
            A :class:`bool` indicating whether the temporal point is ever within `distance` of `other`.

        MEOS Functions:
            edwithin_tpoint_geo, edwithin_tpoint_tpoint
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = edwithin_tpoint_geo(self._inner, gs, distance)
        elif isinstance(other, STBox):
            result = edwithin_tpoint_geo(self._inner, stbox_to_geo(other._inner), distance)
        elif isinstance(other, TPoint):
            result = edwithin_tpoint_tpoint(self._inner, other._inner, distance)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return result == 1

    def ever_intersects(self, other: Union[shpb.BaseGeometry, TPoint, STBox]) -> bool:
        """
        Returns whether the temporal point ever intersects `other`.

        Args:
            other: An object to check for intersection with.

        Returns:
            A :class:`bool` indicating whether the temporal point ever intersects `other`.

        MEOS Functions:
            eintersects_tpoint_geo, eintersects_tpoint_tpoint
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = eintersects_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            result = eintersects_tpoint_geo(self._inner, stbox_to_geo(other._inner))
        elif isinstance(other, TPoint):
            result = eintersects_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return result == 1

    def ever_touches(self, other: Union[shpb.BaseGeometry, STBox]) -> bool:
        """
        Returns whether the temporal point ever touches `other`.

        Args:
            other: An object to check for touching with.

        Returns:
            A :class:`bool` indicating whether the temporal point ever touches `other`.

        MEOS Functions:
            etouches_tpoint_geo
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = etouches_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            result = etouches_tpoint_geo(self._inner, stbox_to_geo(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return result == 1

    # ------------------------- Temporal Spatial Relationships ----------------
    def is_spatially_contained_in(self, container: Union[shpb.BaseGeometry, STBox]) -> TBool:
        """
        Returns a new temporal boolean indicating whether the temporal point is contained by `container`.

        Args:
            container: An object to check for containing `self`.

        Returns:
            A new :TBool: indicating whether the temporal point is contained by `container`.

        MEOS Functions:
            tcontains_geo_tpoint
        """
        from ..boxes import STBox
        if isinstance(container, shpb.BaseGeometry):
            gs = geo_to_gserialized(container, isinstance(self, TGeogPoint))
            result = tcontains_geo_tpoint(gs, self._inner, False, False)
        elif isinstance(container, STBox):
            gs = stbox_to_geo(container._inner)
            result = tcontains_geo_tpoint(gs, self._inner, False, False)
        else:
            raise TypeError(f'Operation not supported with type {container.__class__}')
        return Temporal._factory(result)

    def disjoint(self, other: Union[shpb.BaseGeometry, STBox]) -> TBool:
        """
        Returns a new temporal boolean indicating whether the temporal point intersects `other`.

        Args:
            other: An object to check for intersection with.

        Returns:
            A new :TBool: indicating whether the temporal point intersects `other`.

        MEOS Functions:
            tintersects_tpoint_geo
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = tdisjoint_tpoint_geo(self._inner, gs, False, False)
        elif isinstance(other, STBox):
            result = tdisjoint_tpoint_geo(self._inner, stbox_to_geo(other._inner), False, False)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def within_distance(self, other: Union[shpb.BaseGeometry, TPoint, STBox], distance: float) -> TBool:
        """
        Returns a new temporal boolean indicating whether the temporal point is within `distance` of `other`.

        Args:
            other: An object to check the distance to.
            distance: The distance to check in units of the spatial reference system.

        Returns:
            A new :TBool: indicating whether the temporal point is within `distance` of `other`.

        MEOS Functions:
            tdwithin_tpoint_geo, tdwithin_tpoint_tpoint
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = tdwithin_tpoint_geo(self._inner, gs, distance, False, False)
        elif isinstance(other, STBox):
            result = tdwithin_tpoint_geo(self._inner, stbox_to_geo(other._inner), distance, False, False)
        elif isinstance(other, TPoint):
            result = tdwithin_tpoint_tpoint(self._inner, other._inner, distance, False, False)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def intersects(self, other: Union[shpb.BaseGeometry, STBox]) -> TBool:
        """
        Returns a new temporal boolean indicating whether the temporal point intersects `other`.

        Args:
            other: An object to check for intersection with.

        Returns:
            A new :TBool: indicating whether the temporal point intersects `other`.

        MEOS Functions:
            tintersects_tpoint_geo
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = tintersects_tpoint_geo(self._inner, gs, False, False)
        elif isinstance(other, STBox):
            result = tintersects_tpoint_geo(self._inner, stbox_to_geo(other._inner), False, False)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def touches(self, other: Union[shpb.BaseGeometry, STBox]) -> TBool:
        """
        Returns a new temporal boolean indicating whether the temporal point touches `other`.

        Args:
            other: An object to check for touching with.

        Returns:
            A new :TBool: indicating whether the temporal point touches `other`.

        MEOS Functions:
            ttouches_tpoint_geo
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = ttouches_tpoint_geo(self._inner, gs, False, False)
        elif isinstance(other, STBox):
            result = ttouches_tpoint_geo(self._inner, stbox_to_geo(other._inner), False, False)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other: Union[shpb.BaseGeometry, TPoint, STBox]) -> TFloat:
        """
        Returns the temporal distance between the temporal point and `other`.

        Args:
            other: An object to check the distance to.

        Returns:
            A new :class:`TFloat` indicating the temporal distance between the temporal point and `other`.

        MEOS Functions:
            distance_tpoint_point, distance_tpoint_tpoint
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = distance_tpoint_point(self._inner, gs)
        elif isinstance(other, STBox):
            result = distance_tpoint_point(self._inner, stbox_to_geo(other._inner))
        elif isinstance(other, TPoint):
            result = distance_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def nearest_approach_distance(self, other: Union[shpb.BaseGeometry, STBox, TPoint]) -> float:
        """
        Returns the nearest approach distance between the temporal point and `other`.

        Args:
            other: An object to check the nearest approach distance to.

        Returns:
            A :class:`float` indicating the nearest approach distance between the temporal point and `other`.

        MEOS Functions:
            nad_tpoint_geo, nad_tpoint_stbox, nad_tpoint_tpoint
        """
        from ..boxes import STBox
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            return nad_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return nad_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return nad_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def nearest_approach_instant(self, other: Union[shpb.BaseGeometry, TPoint]) -> TI:
        """
        Returns the nearest approach instant between the temporal point and `other`.

        Args:
            other: An object to check the nearest approach instant to.

        Returns:
            A new temporal instant indicating the nearest approach instant between the temporal point and `other`.

        MEOS Functions:
            nai_tpoint_geo, nai_tpoint_tpoint
        """
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = nai_tpoint_geo(self._inner, gs)
        elif isinstance(other, TPoint):
            result = nai_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def shortest_line(self, other: Union[shpb.BaseGeometry, TPoint]) -> shpb.BaseGeometry:
        """
        Returns the shortest line between the temporal point and `other`.

        Args:
            other: An object to check the shortest line to.

        Returns:
            A new :class:`~shapely.geometry.base.BaseGeometry` indicating the shortest line between the temporal point
            and `other`.

        MEOS Functions:
            shortestline_tpoint_geo, shortestline_tpoint_tpoint
        """
        if isinstance(other, shpb.BaseGeometry):
            gs = geo_to_gserialized(other, isinstance(self, TGeogPoint))
            result = shortestline_tpoint_geo(self._inner, gs)
        elif isinstance(other, TPoint):
            result = shortestline_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return gserialized_to_shapely_geometry(result[0], 10)

    # ------------------------- Tiling Operations -----------------------------
    def tile(self, size: float, duration: Optional[Union[timedelta, str]] = None,
             origin: Optional[shpb.BaseGeometry] = None,
             start: Union[datetime, str, None] = None) -> List[List[List[List[TG]]]]:
        """
        Split the temporal point into segments following the tiling of the
        bounding box.

        Args:
            size: The size of the spatial tiles. If `self` has a spatial
                dimension and this argument is not provided, the tiling will be
                only temporal.
            duration: The duration of the temporal tiles. If `self` has a time 
                dimension and this argument is not provided, the tiling will be
                only spatial.
            origin: The origin of the spatial tiling. If not provided, the 
                origin will be (0, 0, 0).
            start: The start time of the temporal tiling. If not provided, 
                the start time used by default is Monday, January 3, 2000.

        Returns:
            A 4D matrix (XxYxZxT) of :class:`TPoint` objects.

        See Also:
            :meth:`STBox.tile`
        """
        from ..boxes import STBox
        bbox = STBox.from_tpoint(self)
        tiles = bbox.tile(size, duration, origin, start)
        return [[[[self.at(tile) for tile in z_dim]
                  for z_dim in y_dim] for y_dim in x_dim] for x_dim in tiles]

    def tile_flat(self, size: float, duration: Optional[Union[timedelta, str]] = None,
                  origin: Optional[shpb.BaseGeometry] = None,
                  start: Union[datetime, str, None] = None) -> List[TG]:
        """
        Split the temporal point into segments following the tiling of the
        bounding box.

        Args:
            size: The size of the spatial tiles. If `self` has a spatial
                dimension and this argument is not provided, the tiling will be
                only temporal.
            duration: The duration of the temporal tiles. If `self` has a time
                dimension and this argument is not provided, the tiling will be
                only spatial.
            origin: The origin of the spatial tiling. If not provided, the
                origin will be (0, 0, 0).
            start: The start time of the temporal tiling. If not provided, the
                the start time used by default is Monday, January 3, 2000.

        Returns:
            A :class:`list` of :class:`TPoint` objects.

        See Also:
            :meth:`STBox.tile_flat`
        """
        from ..boxes import STBox
        bbox = STBox.from_tpoint(self)
        tiles = bbox.tile_flat(size, duration, origin, start)
        return [x for x in (self.at(tile) for tile in tiles) if x]

    # ------------------------- Split Operations ------------------------------
    def space_split(self, xsize: float, ysize: Optional[float] = None,
                    zsize: Optional[float] = None, origin: Optional[shpb.BaseGeometry] = None,
                    bitmatrix: Optional[bool] = False) -> List[Temporal]:
        """
        Splits `self` into fragments with respect to space buckets

        Args:
            xsize: Size of the x dimension.
            ysize: Size of the y dimension.
            zsize: Size of the z dimension.
            origin: The origin of the spatial tiling. If not provided, the
                origin will be (0, 0, 0).

        Returns:
            A list of temporal points.

        MEOS Functions:
            tpoint_value_split
        """
        ysz = ysize if ysize is not None else xsize
        zsz = zsize if zsize is not None else xsize
        gs = geo_to_gserialized(origin, self.geodetic()) if origin is not None \
            else pgis_geography_in('Point(0 0 0)', -1) \
            if isinstance(self, TGeogPoint) \
            else pgis_geometry_in('Point(0 0 0)', -1)
        fragments, values, count = tpoint_space_split(self._inner,
                                                      xsize, ysz, zsz, gs, bitmatrix)
        from ..factory import _TemporalFactory
        return [_TemporalFactory.create_temporal(fragments[i]) for i in \
                range(count)]

    def space_time_split(self, xsize: float, duration: Union[str, timedelta],
                         ysize: Optional[float] = None, zsize: Optional[float] = None,
                         origin: Optional[shpb.BaseGeometry] = None,
                         time_start: Optional[Union[str, datetime]] = None,
                         bitmatrix: Optional[bool] = False) -> List[Temporal]:
        """
        Splits `self` into fragments with respect to space and period buckets.

        Args:
            xsize: Size of the x dimension.
            ysize: Size of the y dimension.
            zsize: Size of the z dimension.
            duration: Duration of the period buckets.
            origin: The origin of the spatial tiling. If not provided, the
                origin will be (0, 0, 0).
            time_start: Start time of the first period bucket. If None, the
                start time used by default is Monday, January 3, 2000.

        Returns:
            A list of temporal floats.

        MEOS Functions:
            tfloat_value_time_split
        """
        ysz = ysize if ysize is not None else xsize
        zsz = zsize if zsize is not None else xsize
        dt = timedelta_to_interval(duration) \
            if isinstance(duration, timedelta) \
            else pg_interval_in(duration, -1)
        gs = geo_to_gserialized(origin, self.geodetic()) if origin is not None \
            else pgis_geography_in('Point(0 0 0)', -1) \
            if isinstance(self, TGeogPoint) \
            else pgis_geometry_in('Point(0 0 0)', -1)
        if time_start is None:
            st = pg_timestamptz_in('2000-01-03', -1)
        else:
            st = datetime_to_timestamptz(time_start) \
                if isinstance(time_start, datetime) \
                else pg_timestamptz_in(time_start, -1)
        fragments, points, times, count = tpoint_space_time_split(self._inner,
                                                                  xsize, ysz, zsz, dt, gs, st, bitmatrix)
        return [Temporal._factory(fragments[i]) for i in range(count)]


class TPointInst(TInstant[shpb.BaseGeometry, TG, TI, TS, TSS], TPoint[TG, TI, TS, TSS], ABC):
    """
    Abstract class for temporal point instants.
    """

    def value(self, precision: int = 15) -> shp.Point:
        """
        Returns the value of the temporal point instant.

        Args:
            precision: The precision of the returned geometry.

        Returns:
            A new :class:`~shapely.geometry.point.Point` representing the value of the temporal point instant.
        """
        return self.start_value(precision=precision)

    def x(self) -> TFloatInst:
        return super().x()

    def y(self) -> TFloatInst:
        return super().y()

    def z(self) -> TFloatInst:
        return super().z()


class TPointSeq(TSequence[shpb.BaseGeometry, TG, TI, TS, TSS], TPoint[TG, TI, TS, TSS], ABC):
    """
    Abstract class for temporal point sequences.
    """

    def x(self) -> TFloatSeq:
        return super().x()

    def y(self) -> TFloatSeq:
        return super().y()

    def z(self) -> TFloatSeq:
        return super().z()

    def plot(self, *args, **kwargs):
        """
        Plots the temporal point sequence.

        Args:
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.

        See Also:
            :func:`~pygeos.plotters.point_sequence_plotter.TemporalPointSequencePlotter.plot_xy`
        """
        from ..plotters import TemporalPointSequencePlotter
        return TemporalPointSequencePlotter.plot_xy(self, *args, **kwargs)


class TPointSeqSet(TSequenceSet[shpb.BaseGeometry, TG, TI, TS, TSS], TPoint[TG, TI, TS, TSS], ABC):
    """
    Abstract class for temporal point sequence sets.
    """

    def to_dataframe(self, precision: int = 15) -> GeoDataFrame:
        """
        Returns the temporal point sequence set as a GeoPandas DataFrame.

        Args:
            precision: The precision of the returned geometry.

        Returns:
            A new :class:`GeoDataFrame` representing the temporal point
            sequence set.
        """
        gpd = import_geopandas()
        sequences = self.sequences()
        data = {
            'sequence': [i + 1 for i, seq in enumerate(sequences) for _ in
                         range(seq.num_instants())],
            'time': [t for seq in sequences for t in seq.timestamps()],
            'geometry': [v for seq in sequences for v in
                         seq.values(precision=precision)]
        }
        return gpd.GeoDataFrame(data, crs=self.srid()).set_index(keys=['sequence', 'time'])

    def x(self) -> TFloatSeqSet:
        return super().x()

    def y(self) -> TFloatSeqSet:
        return super().y()

    def z(self) -> TFloatSeqSet:
        return super().z()

    def plot(self, *args, **kwargs):
        """
        Plots the temporal point sequence set.

        Args:
            *args: Additional arguments to pass to the plot function.
            **kwargs: Additional keyword arguments to pass to the plot function.

        Returns:
            List with the plotted elements.

        See Also:
            :func:`~pygeos.plotters.point_sequenceset_plotter.TemporalPointSequenceSetPlotter.plot_xy`
        """
        from ..plotters import TemporalPointSequenceSetPlotter
        return TemporalPointSequenceSetPlotter.plot_xy(self, *args, **kwargs)


class TGeomPoint(TPoint['TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq',
'TGeomPointSeqSet'], ABC):
    """
    Abstract class for temporal geometric points.
    """

    _mobilitydb_name = 'tgeompoint'

    BaseClass = shp.Point
    _parse_function = tgeompoint_in

    # ------------------------- Output ----------------------------------------
    @staticmethod
    def from_base_temporal(value: shpb.BaseGeometry,
                           base: Temporal) -> TGeomPoint:
        """
        Creates a temporal geometric point from a base geometry and the time
        frame of another temporal object.

        Args:
            value: The base geometry.
            base: The temporal object defining the time frame.

        Returns:
            A new :class:`TGeomPoint` object.

        MEOS Functions:
            tpoint_from_base_temp
        """
        gs = geometry_to_gserialized(value)
        result = tpoint_from_base_temp(gs, base._inner)
        return Temporal._factory(result)

    @staticmethod
    @overload
    def from_base_time(value: shpb.BaseGeometry,
                       base: datetime) -> TGeomPointInst:
        ...

    @staticmethod
    @overload
    def from_base_time(value: shpb.BaseGeometry,
                       base: Union[TimestampSet, Period]) -> TGeomPointSeq:
        ...

    @staticmethod
    @overload
    def from_base_time(value: shpb.BaseGeometry,
                       base: PeriodSet) -> TGeomPointSeqSet:
        ...

    @staticmethod
    def from_base_time(value: shpb.BaseGeometry, base: Time,
                       interpolation: TInterpolation = None) -> TGeomPoint:
        """
        Creates a temporal geometric point from a base geometry and a time value.

        Args:
            value: The base geometry.
            base: The time value.
            interpolation: The interpolation method.

        Returns:
            A new :class:`TGeomPoint` object.

        MEOS Functions:
            tpointinst_make, tpointseq_from_base_timestampset,
            tpointseq_from_base_period, tpointseqset_from_base_periodset
        """
        gs = geometry_to_gserialized(value)
        if isinstance(base, datetime):
            return TGeomPointInst(_inner=tpointinst_make(gs,
                                                         datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TGeomPointSeq(_inner=tpointseq_from_base_timestampset(gs,
                                                                         base._inner))
        elif isinstance(base, Period):
            return TGeomPointSeq(_inner=tpointseq_from_base_period(gs,
                                                                   base._inner, interpolation))
        elif isinstance(base, PeriodSet):
            return TGeomPointSeqSet(_inner=tpointseqset_from_base_periodset(gs,
                                                                            base._inner, interpolation))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    # ------------------------- Conversions ----------------------------------
    def to_geographic(self) -> TGeogPoint:
        """
        Returns a copy of `self` converted to geographic coordinates.

        Returns:
            A new :class:`TGeogPoint` object.

        MEOS Functions:
            tgeompoint_to_tgeogpoint
        """
        result = tgeompoint_to_tgeogpoint(self._inner)
        return Temporal._factory(result)

    def to_dataframe(self) -> GeoDataFrame:
        """
        Returns the trajectory of the temporal point as a GeoPandas DataFrame.

        Returns:
            A new :class:`GeoDataFrame` representing the trajectory.
        """
        gpd = import_geopandas()
        data = {
            'time': self.timestamps(),
            'geometry': [i.value() for i in self.instants()]
        }
        return gpd.GeoDataFrame(data, crs=self.srid()).set_index(keys=['time'])

    # ------------------------- Ever and Always Comparisons -------------------
    def always_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is always equal to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is always equal to `value`, False otherwise.

        MEOS Functions:
            tpoint_always_eq
        """
        gs = geometry_to_gserialized(value)
        return tpoint_always_eq(self._inner, gs)

    def always_not_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is always different to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is always different to `value`, False otherwise.

        MEOS Functions:
            tpoint_ever_eq
        """
        gs = geometry_to_gserialized(value)
        return not tpoint_ever_eq(self._inner, gs)

    def ever_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is ever equal to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is ever equal to `value`, False otherwise.

        MEOS Functions:
            tpoint_ever_eq
        """
        gs = geometry_to_gserialized(value)
        return tpoint_ever_eq(self._inner, gs)

    def ever_not_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is ever different to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is ever different to `value`, False otherwise.

        MEOS Functions:
            tpoint_always_eq
        """
        gs = geometry_to_gserialized(value)
        return not tpoint_always_eq(self._inner, gs)

    def never_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is never equal to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is never equal to `value`, False otherwise.

        MEOS Functions:
            tpoint_ever_eq
        """
        gs = geometry_to_gserialized(value)
        return not tpoint_ever_eq(self._inner, gs)

    def never_not_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is never different to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is never different to `value`, False otherwise.

        MEOS Functions:
            tpoint_always_eq
        """
        gs = geometry_to_gserialized(value)
        return tpoint_always_eq(self._inner, gs)

    # ------------------------- Temporal Comparisons --------------------------
    def temporal_equal(self, other: Union[shp.Point, Temporal]) -> TBool:
        """
        Returns the temporal equality relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal equality relation.

        MEOS Functions:
            teq_tpoint_point, teq_temporal_temporal
        """
        if isinstance(other, shp.Point):
            gs = geometry_to_gserialized(other)
            result = teq_tpoint_point(self._inner, gs)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[shp.Point, Temporal]) -> Temporal:
        """
        Returns the temporal inequality relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal inequality relation.

        MEOS Functions:
            tne_tpoint_point, tne_temporal_temporal
        """
        if isinstance(other, shp.Point):
            gs = geometry_to_gserialized(other)
            result = tne_tpoint_point(self._inner, gs)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    # ------------------------- Database Operations ---------------------------
    @staticmethod
    def read_from_cursor(value, _=None):
        """
        Reads a :class:`TGeogPoint` from a database cursor. Used when
        automatically loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        if value.startswith('Interp=Stepwise;'):
            value1 = value.replace('Interp=Stepwise;', '')
            if value1[0] == '{':
                return TGeomPointSeqSet(string=value)
            else:
                return TGeomPointSeq(string=value)
        elif value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TGeomPointInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TGeomPointSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TGeomPointSeqSet(string=value)
            else:
                return TGeomPointSeq(string=value)
        raise Exception("ERROR: Could not parse temporal point value")


class TGeogPoint(TPoint['TGeogPoint', 'TGeogPointInst', 'TGeogPointSeq',
'TGeogPointSeqSet'], ABC):
    """
    Abstract class for representing temporal geographic points.
    """

    _mobilitydb_name = 'tgeogpoint'

    BaseClass = shp.Point
    _parse_function = tgeogpoint_in

    # ------------------------- Output ----------------------------------------
    @staticmethod
    def from_base_temporal(value: shpb.BaseGeometry,
                           base: Temporal) -> TGeogPoint:
        """
        Creates a temporal geographic point from a base geometry and the time
        frame of another temporal object.

        Args:
            value: The base geometry.
            base: The temporal object defining the time frame.

        Returns:
            A new :class:`TGeogPoint` object.

        MEOS Functions:
            tpoint_from_base_temp
        """
        gs = geography_to_gserialized(value)
        result = tpoint_from_base_temp(gs, base._inner)
        return Temporal._factory(result)

    @staticmethod
    @overload
    def from_base_time(value: shpb.BaseGeometry,
                       base: datetime) -> TGeogPointInst:
        ...

    @staticmethod
    @overload
    def from_base_time(value: shpb.BaseGeometry,
                       base: Union[TimestampSet, Period]) -> TGeogPointSeq:
        ...

    @staticmethod
    @overload
    def from_base_time(value: shpb.BaseGeometry,
                       base: PeriodSet) -> TGeogPointSeqSet:
        ...

    @staticmethod
    def from_base_time(value: shpb.BaseGeometry, base: Time,
                       interpolation: TInterpolation = None) -> TGeogPoint:
        """
        Creates a temporal geographic point from a base geometry and a time object.

        Args:
            value: The base geometry.
            base: The time object defining the time frame.
            interpolation: The interpolation method.

        Returns:
            A new :class:`TGeogPoint` object.

        MEOS Functions:
            tpointinst_make, tpointseq_from_base_timestampset,
            tpointseq_from_base_period, tpointseqset_from_base_periodset
        """
        gs = geography_to_gserialized(value)
        if isinstance(base, datetime):
            return TGeogPointInst(_inner=tpointinst_make(gs,
                                                         datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TGeogPointSeq(_inner=tpointseq_from_base_timestampset(gs,
                                                                         base._inner))
        elif isinstance(base, Period):
            return TGeogPointSeq(_inner=tpointseq_from_base_period(gs,
                                                                   base._inner, interpolation))
        elif isinstance(base, PeriodSet):
            return TGeogPointSeqSet(_inner=tpointseqset_from_base_periodset(gs,
                                                                            base._inner, interpolation))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    # ------------------------- Conversions ----------------------------------
    def to_geometric(self) -> TGeomPoint:
        """
        Converts the temporal geographic point to a temporal geometric point.

        Returns:
            A new :class:`TGeomPoint` object.

        MEOS Functions:
            tgeogpoint_to_tgeompoint
        """
        result = tgeogpoint_to_tgeompoint(self._inner)
        return Temporal._factory(result)

    # ------------------------- Ever and Always Comparisons -------------------
    def always_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is always equal to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is always equal to `value`, False otherwise.

        MEOS Functions:
            tpoint_always_eq
        """
        gs = geography_to_gserialized(value)
        return tpoint_always_eq(self._inner, gs)

    def always_not_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is always different to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is always different to `value`, False otherwise.

        MEOS Functions:
            tpoint_ever_eq
        """
        gs = geography_to_gserialized(value)
        return not tpoint_ever_eq(self._inner, gs)

    def ever_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is ever equal to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is ever equal to `value`, False otherwise.

        MEOS Functions:
            tpoint_ever_eq
        """
        gs = geography_to_gserialized(value)
        return tpoint_ever_eq(self._inner, gs)

    def ever_not_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is ever different to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is ever different to `value`, False otherwise.

        MEOS Functions:
            tpoint_always_eq
        """
        gs = geography_to_gserialized(value)
        return not tpoint_always_eq(self._inner, gs)

    def never_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is never equal to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is never equal to `value`, False otherwise.

        MEOS Functions:
            tpoint_ever_eq
        """
        gs = geography_to_gserialized(value)
        return not tpoint_ever_eq(self._inner, gs)

    def never_not_equal(self, value: shpb.BaseGeometry) -> bool:
        """
        Returns whether `self` is never different to `value`.

        Args:
            value: The geometry to compare with.

        Returns:
            True if `self` is never different to `value`, False otherwise.

        MEOS Functions:
            tpoint_always_eq
        """
        gs = geography_to_gserialized(value)
        return tpoint_always_eq(self._inner, gs)

    # ------------------------- Temporal Comparisons --------------------------
    def temporal_equal(self, other: Union[shp.Point, Temporal]) -> TBool:
        """
        Returns the temporal equality relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal equality relation.

        MEOS Functions:
            teq_tpoint_point, teq_temporal_temporal
        """
        if isinstance(other, shp.Point):
            gs = geography_to_gserialized(other)
            result = teq_tpoint_point(self._inner, gs)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[shp.Point, Temporal]) -> TBool:
        """
        Returns the temporal inequality relation between `self` and `other`.

        Args:
            other: A temporal object to compare to `self`.

        Returns:
            A :class:`TBool` with the result of the temporal inequality relation.

        MEOS Functions:
            tne_tpoint_point, tne_temporal_temporal
        """
        if isinstance(other, shp.Point):
            gs = geography_to_gserialized(other)
            result = tne_tpoint_point(self._inner, gs)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    # ------------------------- Database Operations ---------------------------
    @staticmethod
    def read_from_cursor(value, _=None):
        """
        Reads a :class:`TGeogPoint` from a database cursor. Used when
        automatically loading objects from the database.
        Users should use the class constructor instead.
        """
        if not value:
            return None
        if value.startswith('Interp=Stepwise;'):
            value1 = value.replace('Interp=Stepwise;', '')
            if value1[0] == '{':
                return TGeogPointSeqSet(string=value)
            else:
                return TGeogPointSeq(string=value)
        elif value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TGeogPointInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TGeogPointSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TGeogPointSeqSet(string=value)
            else:
                return TGeomPointSeq(string=value)
        raise Exception("ERROR: Could not parse temporal point value")


class TGeomPointInst(TPointInst['TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq', 'TGeomPointSeqSet'], TGeomPoint):
    """
    Class for representing temporal geometric points at a single instant.
    """
    _make_function = lambda *args: None
    _cast_function = lambda x: None

    def __init__(self, string: Optional[str] = None, *,
                 point: Optional[Union[str, shp.Point, Tuple[float, float], Tuple[float, float, float]]] = None,
                 timestamp: Optional[Union[str, datetime]] = None,
                 srid: Optional[int] = 0, _inner=None) -> None:
        super().__init__(string=string, value=point, timestamp=timestamp,
                         _inner=_inner)
        if self._inner is None:
            if isinstance(point, tuple):
                p = f'POINT({" ".join(str(x) for x in point)})'
            else:
                p = f'{point}'
            self._inner = tgeompoint_in(f"SRID={srid};{p}@{timestamp}")


class TGeogPointInst(TPointInst['TGeogPoint', 'TGeogPointInst',
'TGeogPointSeq', 'TGeogPointSeqSet'], TGeogPoint):
    """
    Class for representing temporal geographic points at a single instant.
    """
    _make_function = lambda *args: None
    _cast_function = lambda x: None

    def __init__(self, string: Optional[str] = None, *,
                 point: Optional[Union[str, shp.Point, Tuple[float, float], Tuple[float, float, float]]] = None,
                 timestamp: Optional[Union[str, datetime]] = None,
                 srid: Optional[int] = 4326, _inner=None) -> None:
        super().__init__(string=string, value=point, timestamp=timestamp,
                         _inner=_inner)
        if self._inner is None:
            p = f'POINT({point[0]} {point[1]})' if isinstance(point, tuple) \
                else f'{point}'
            self._inner = tgeogpoint_in(f"SRID={srid};{p}@{timestamp}")


class TGeomPointSeq(TPointSeq['TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq',
'TGeomPointSeqSet'], TGeomPoint):
    """
    Class for representing temporal geometric points over a period of time.
    """
    ComponentClass = TGeomPointInst

    def __init__(self, string: Optional[str] = None, *,
                 instant_list: Optional[List[Union[str, TGeomPointInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False,
                 interpolation: TInterpolation = TInterpolation.LINEAR,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list,
                         lower_inc=lower_inc, upper_inc=upper_inc,
                         interpolation=interpolation,
                         normalize=normalize, _inner=_inner)


class TGeogPointSeq(TPointSeq['TGeogPoint', 'TGeogPointInst', 'TGeogPointSeq',
'TGeogPointSeqSet'], TGeogPoint):
    """
    Class for representing temporal geographic points over a period of time.
    """
    ComponentClass = TGeogPointInst

    def __init__(self, string: Optional[str] = None, *,
                 instant_list: Optional[List[Union[str, TGeogPointInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False,
                 interpolation: TInterpolation = TInterpolation.LINEAR,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list,
                         lower_inc=lower_inc, upper_inc=upper_inc,
                         interpolation=interpolation,
                         normalize=normalize, _inner=_inner)


class TGeomPointSeqSet(TPointSeqSet['TGeomPoint', 'TGeomPointInst',
'TGeomPointSeq', 'TGeomPointSeqSet'], TGeomPoint):
    """
    Class for representing temporal geometric points over a period of time
    with gaps.
    """
    ComponentClass = TGeomPointSeq

    def __init__(self, string: Optional[str] = None, *,
                 sequence_list: Optional[List[Union[str, TGeomPointSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list,
                         normalize=normalize, _inner=_inner)


class TGeogPointSeqSet(TPointSeqSet['TGeogPoint', 'TGeogPointInst',
'TGeogPointSeq', 'TGeogPointSeqSet'], TGeogPoint):
    """
    Class for representing temporal geographic points over a period of time
    with gaps.
    """
    ComponentClass = TGeogPointSeq

    def __init__(self, string: Optional[str] = None, *,
                 sequence_list: Optional[List[Union[str, TGeogPointSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list,
                         normalize=normalize,
                         _inner=_inner)



from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..boxes import STBox
from ..main import TPoint


class TemporalPointExtentAggregator(BaseAggregator[TPoint, STBox]):
    """
    Spatiotemporal extent of aggregated temporal points, i.e. smallest :class:`~pymeos.time.boxes.STBox` that
    includes all aggregated temporal numbers.

    MEOS Functions:
        tpoint_extent_transfn, temporal_tagg_finalfn
    """
    _add_function = tpoint_extent_transfn

    @classmethod
    def _finish(cls, state) -> STBox:
        return STBox(_inner=state)



from .interpolation import TInterpolation
from .temporal import Temporal
from .tinstant import TInstant
from .tsequence import TSequence
from .tsequenceset import TSequenceSet

__all__ = ['TInterpolation', 'Temporal', 'TInstant', 'TSequence', 'TSequenceSet']



from __future__ import annotations

from typing import List, Union, overload, Optional, TYPE_CHECKING

from pymeos_cffi import intset_in, intset_make, intset_out, intset_start_value, \
    intset_end_value, intset_value_n, intset_values, contains_intset_int, \
    intersection_intset_int, intersection_set_set, minus_intset_int, \
    left_intset_int, overleft_intset_int, right_intset_int, overright_intset_int, \
    minus_set_set, union_set_set, union_intset_int, intset_shift_scale, \
    minus_int_intset, distance_intset_int

from .intspan import IntSpan
from .intspanset import IntSpanSet
from ..base import Set

if TYPE_CHECKING:
    from .floatset import FloatSet


class IntSet(Set[int]):
    """
    Class for representing a set of text values.

    ``TextSet`` objects can be created with a single argument of type string as
    in MobilityDB.

        >>> IntSet(string='{1, 3, 56}')

    Another possibility is to create a ``TextSet`` object from a list of
    strings or integers.

        >>> IntSet(elements=[1, '2', 3, '56'])


    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'intset'

    _parse_function = intset_in
    _parse_value_function = int
    _make_function = intset_make

    # ------------------------- Constructors ----------------------------------

    # ------------------------- Output ----------------------------------------

    def __str__(self):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            intset_out
        """
        return intset_out(self._inner)

    # ------------------------- Conversions -----------------------------------

    def to_spanset(self) -> IntSpanSet:
        """
        Returns a SpanSet that contains a Span for each element in ``self``.

        Returns:
            A new :class:`IntSpanSet` instance

        MEOS Functions:
            set_to_spanset
        """

        return IntSpanSet(_inner=super().to_spanset())

    def to_span(self) -> IntSpan:
        """
        Returns a span that encompasses ``self``.

        Returns:
            A new :class:`IntSpan` instance

        MEOS Functions:
            set_span
        """
        return IntSpan(_inner=super().to_span())

    def to_floatset(self) -> FloatSet:
        """
        Converts ``self`` to a :class:`FloatSet` instance.

        Returns:
            A new :class:`FloatSet` instance
        """
        from .floatset import FloatSet
        return FloatSet(elements=[float(x) for x in self.elements()])

    # ------------------------- Accessors -------------------------------------

    def start_element(self) -> int:
        """
        Returns the first element in ``self``.

        Returns:
            A :class:`int` instance

        MEOS Functions:
            intset_start_value
        """
        return intset_start_value(self._inner)

    def end_element(self) -> int:
        """
        Returns the last element in ``self``.

        Returns:
            A :class:`int` instance

        MEOS Functions:
            intset_end_value
        """
        return intset_end_value(self._inner)

    def element_n(self, n: int) -> int:
        """
        Returns the ``n``-th element in ``self``.

        Args:
            n: The 0-based index of the element to return.

        Returns:
            A :class:`int` instance

        MEOS Functions:
            intset_value_n
        """
        super().element_n(n)
        return intset_value_n(self._inner, n + 1)

    def elements(self) -> List[int]:
        """
        Returns the elements in ``self``.

        Returns:
            A list of :class:`int` instances

        MEOS Functions:
            inttset_values
        """
        elems = intset_values(self._inner)
        return [elems[i] for i in range(self.num_elements())]

    # ------------------------- Transformations ------------------------------------

    def shift(self, delta: int) -> IntSet:
        """
        Returns a new ``IntSet`` instance with all elements shifted by ``delta``.

        Args:
            delta: The value to shift by.

        Returns:
            A new :class:`IntSet` instance

        MEOS Functions:
            intset_shift_scale
        """
        return self.shift_scale(delta, None)

    def scale(self, width: int) -> IntSet:
        """
        Returns a new ``IntSet`` instance with all elements scaled to so that the encompassing
        span has width ``width``.

        Args:
            width: The new width.

        Returns:
            A new :class:`IntSet` instance

        MEOS Functions:
            intset_shift_scale
        """
        return self.shift_scale(None, width)

    def shift_scale(self, delta: Optional[int], width: Optional[int]) -> IntSet:
        """
        Returns a new ``IntSet`` instance with all elements shifted by
        ``delta`` and scaled to so that the encompassing span has width
        ``width``.

        Args:
            delta: The value to shift by.
            width: The new width.

        Returns:
            A new :class:`IntSet` instance

        MEOS Functions:
            intset_shift_scale
        """
        return IntSet(
            _inner=intset_shift_scale(self._inner, delta, width,
                                      delta is not None, width is not None))

    # ------------------------- Topological Operations --------------------------------

    def contains(self, content: Union[IntSet, int]) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_set_set, contains_intset_int
        """
        if isinstance(content, int):
            return contains_intset_int(self._inner, content)
        else:
            return super().contains(content)

    # ------------------------- Position Operations --------------------------------

    def is_left(self, content: Union[IntSet, int]) -> bool:
        """
        Returns whether ``self`` is strictly to the left of ``other``. That is,
        ``self`` ends before ``other`` starts.

        Args:
            content: object to compare with

        Returns:
            True if left, False otherwise

        MEOS Functions:
            left_set_set, left_intset_int
        """
        if isinstance(content, int):
            return left_intset_int(self._inner, content)
        else:
            return super().is_left(content)

    def is_over_or_left(self, content: Union[IntSet, int]) -> bool:
        """
        Returns whether ``self`` is to the left of ``other`` allowing overlap.
        That is, ``self`` ends before ``other`` ends (or at the same value).

        Args:
            content: object to compare with

        Returns:
            True if is over or left, False otherwise

        MEOS Functions:
            overleft_set_set, overleft_intset_int
        """
        if isinstance(content, int):
            return overleft_intset_int(self._inner, content)
        else:
            return super().is_over_or_left(content)

    def is_right(self, content: Union[IntSet, int]) -> bool:
        """
        Returns whether ``self`` is strictly to the right of ``other``. That is,
        ``self`` ends after ``other`` starts.

        Args:
            content: object to compare with

        Returns:
            True if right, False otherwise

        MEOS Functions:
            right_set_set, right_intset_int
        """
        if isinstance(content, int):
            return right_intset_int(self._inner, content)
        else:
            return super().is_right(content)

    def is_over_or_right(self, content: Union[IntSet, int]) -> bool:
        """
        Returns whether ``self`` is to the right of ``other`` allowing overlap.
        That is, ``self`` starts before ``other`` ends (or at the same value).

        Args:
            content: object to compare with

        Returns:
            True if is over or right, False otherwise

        MEOS Functions:
            overright_set_set, overright_intset_int
        """
        if isinstance(content, int):
            return overright_intset_int(self._inner, content)
        else:
            return super().is_over_or_right(content)

    # ------------------------- Set Operations --------------------------------

    @overload
    def intersection(self, other: int) -> Optional[int]:
        ...

    @overload
    def intersection(self, other: IntSet) -> Optional[IntSet]:
        ...

    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: A :class:`IntSet` or :class:`int` instance

        Returns:
            An object of the same type as ``other`` or ``None`` if the
            intersection is empty.

        MEOS Functions:
            intersection_set_set, intersection_intset_int
        """
        if isinstance(other, int):
            return intersection_intset_int(self._inner, other)
        elif isinstance(other, IntSet):
            result = intersection_set_set(self._inner, other._inner)
            return IntSet(_inner=result) if result is not None else None
        else:
            return super().intersection(other)

    def minus(self, other: Union[IntSet, int]) -> Optional[IntSet]:
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: A :class:`IntSet` or :class:`int` instance

        Returns:
            A :class:`IntSet` instance or ``None`` if the difference is empty.

        MEOS Functions:
            minus_set_set, minus_intset_int
        """
        if isinstance(other, int):
            result = minus_intset_int(self._inner, other)
            return IntSet(_inner=result) if result is not None else None
        elif isinstance(other, IntSet):
            result = minus_set_set(self._inner, other._inner)
            return IntSet(_inner=result) if result is not None else None
        else:
            return super().minus(other)

    def subtract_from(self, other: int) -> Optional[int]:
        """
        Returns the difference of ``other`` and ``self``.

        Args:
            other: A :class:`int` instance

        Returns:
            A :class:`int` instance or ``None`` if the difference is empty.

        MEOS Functions:
            minus_int_intset

        See Also:
            :meth:`minus`
        """
        return minus_int_intset(other, self._inner)

    def union(self, other: Union[IntSet, int]) -> IntSet:
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: A :class:`IntSet` or :class:`int` instance

        Returns:
            A :class:`IntSet` instance.

        MEOS Functions:
            union_set_set, union_intset_int
        """
        if isinstance(other, int):
            result = union_intset_int(self._inner, other)
            return IntSet(_inner=result) if result is not None else None
        elif isinstance(other, IntSet):
            result = union_set_set(self._inner, other._inner)
            return IntSet(_inner=result) if result is not None else None
        else:
            return super().union(other)

    # ------------------------- Distance Operations ---------------------------

    def distance(self, other: Union[int, IntSet, IntSpan, IntSpanSet]) -> float:
        if isinstance(other, int):
            return distance_intset_int(self._inner, other)
        else:
            return super().distance(other)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Union
from typing import TypeVar, Type, Callable, Any, TYPE_CHECKING

from pymeos_cffi import *

from .collection import Collection

if TYPE_CHECKING:
    from .spanset import SpanSet

T = TypeVar('T')
Self = TypeVar('Self', bound='Span[Any]')


class Span(Collection[T], ABC):
    """
    Base class for all span classes.
    """

    __slots__ = ['_inner']

    _parse_function: Callable[[str], 'CData'] = None
    _parse_value_function: Callable[[Union[str, T]], Any] = None
    _make_function: Callable[[Any, Any, bool, bool], 'CData'] = None

    # ------------------------- Constructors ----------------------------------
    def __init__(self, string: Optional[str] = None, *,
                 lower: Optional[Union[str, T]] = None,
                 upper: Optional[Union[str, T]] = None,
                 lower_inc: Optional[bool] = True,
                 upper_inc: Optional[bool] = False,
                 _inner=None):
        super().__init__()
        assert (_inner is not None) or ((string is not None) != \
                                        (lower is not None and upper is not None)), \
            "Either string must be not None or both lower and upper must be not"
        if _inner is not None:
            self._inner = _inner
        elif string is not None:
            self._inner = self.__class__._parse_function(string)
        else:
            lower_converted = self.__class__._parse_value_function(lower)
            upper_converted = self.__class__._parse_value_function(upper)
            self._inner = self.__class__._make_function(lower_converted,
                                                        upper_converted, lower_inc, upper_inc)

    def __copy__(self: Self) -> Self:
        """
        Return a copy of ``self``.

        Returns:
            A new :class:`Span` instance

        MEOS Functions:
            span_copy
        """
        inner_copy = span_copy(self._inner)
        return self.__class__(_inner=inner_copy)

    @classmethod
    def from_wkb(cls: Type[Self], wkb: bytes) -> Self:
        """
        Returns a `Period` from its WKB representation.

        Args:
            wkb: The WKB string.

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            span_from_wkb
        """
        return cls(_inner=(span_from_wkb(wkb)))

    @classmethod
    def from_hexwkb(cls: Type[Self], hexwkb: str) -> Self:
        """
        Returns a `Period` from its WKB representation in hex-encoded ASCII.

        Args:
            hexwkb: WKB representation in hex-encoded ASCII

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            span_from_hexwkb
        """
        result = span_from_hexwkb(hexwkb)
        return cls(_inner=result)

    # ------------------------- Output ----------------------------------------
    @abstractmethod
    def __str__(self) -> str:
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance
        """
        raise NotImplementedError()

    def __repr__(self):
        """
        Return the string representation of ``self``.

        Returns:
            A new :class:`str` instance
        """
        return (f'{self.__class__.__name__}'
                f'({self})')

    def as_wkb(self) -> bytes:
        """
        Returns the WKB representation of ``self``.

        Returns:
            A :class:`str` object with the WKB representation of ``self``.

        MEOS Functions:
            span_as_wkb
        """
        return span_as_wkb(self._inner, 4)

    def as_hexwkb(self) -> str:
        """
        Returns the WKB representation of ``self`` in hex-encoded ASCII.

        Returns:
            A :class:`str` object with the WKB representation of ``self`` in
            hex-encoded ASCII.

        MEOS Functions:
            span_as_hexwkb
        """
        return span_as_hexwkb(self._inner, -1)[0]

    # ------------------------- Conversions -----------------------------------
    @abstractmethod
    def to_spanset(self) -> SpanSet:
        """
        Returns a period set containing ``self``.

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            span_to_spanset
        """
        return span_to_spanset(self._inner)

    # ------------------------- Accessors -------------------------------------
    @abstractmethod
    def lower(self) -> T:
        """
        Returns the lower bound of a period
        """
        return NotImplementedError()

    @abstractmethod
    def upper(self) -> T:
        """
        Returns the upper bound of a period
        """
        return NotImplementedError()

    def lower_inc(self) -> bool:
        """
        Returns whether the lower bound of the period is inclusive or not

        Returns:
            True if the lower bound of the period is inclusive and False
            otherwise

        MEOS Functions:
            span_lower_inc
        """
        return span_lower_inc(self._inner)

    def upper_inc(self) -> bool:
        """
        Returns whether the upper bound of the period is inclusive or not

        Returns:
            True if the upper bound of the period is inclusive and False 
            otherwise

        MEOS Functions:
            span_upper_inc
        """
        return span_upper_inc(self._inner)

    def width(self) -> float:
        """
        Returns the duration of the period.

        Returns:
            Returns a `float` representing the duration of the period in seconds

        MEOS Functions:
            span_width
        """
        return span_width(self._inner)

    def __hash__(self) -> int:
        """
        Return the hash representation of ``self``.

        Returns:
            A new :class:`int` instance

        MEOS Functions:
            span_hash
        """
        return span_hash(self._inner)

    # ------------------------- Topological Operations ------------------------
    @abstractmethod
    def is_adjacent(self: Self, other) -> bool:
        """
        Returns whether ``self`` is adjacent to ``other``. That is, they share
        a bound but only one of them contains it.

        Args:
            other: object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_span_span, adjacent_span_spanset,
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return adjacent_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return adjacent_spanset_span(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_contained_in(self, container) -> bool:
        """
        Returns whether ``self`` is contained in ``container``.

        Args:
            container: temporal object to compare with

        Returns:
            True if contained, False otherwise

        MEOS Functions:
            contained_span_span, contained_span_spanset, contained_period_temporal
        """
        from .spanset import SpanSet
        if isinstance(container, Span):
            return contained_span_span(self._inner, container._inner)
        elif isinstance(container, SpanSet):
            return contained_span_spanset(self._inner, container._inner)
        else:
            raise TypeError(f'Operation not supported with type {container.__class__}')

    def contains(self, content) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_span_span, contains_span_spanset, contains_period_timestamp,
            contains_period_timestampset, contains_period_temporal
        """
        from .spanset import SpanSet
        if isinstance(content, Span):
            return contains_span_span(self._inner, content._inner)
        elif isinstance(content, SpanSet):
            return contains_span_spanset(self._inner, content._inner)
        else:
            raise TypeError(f'Operation not supported with type {content.__class__}')

    def __contains__(self, item):
        """
        Return whether ``self`` contains ``item``.

        Args:
            item: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_span_span, contains_span_spanset, contains_period_timestamp,
            contains_period_timestampset, contains_period_temporal
        """
        return self.contains(item)

    def overlaps(self, other) -> bool:
        """
        Returns whether ``self`` overlaps ``other``. That is, both share at
        least an element.

        Args:
            other: temporal object to compare with

        Returns:
            True if overlaps, False otherwise

        MEOS Functions:
            overlaps_span_span, overlaps_span_spanset
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return overlaps_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return overlaps_spanset_span(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_same(self, other) -> bool:
        """
        Returns whether ``self`` and the bounding period of ``other`` is the
        same.

        Args:
            other: temporal object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            same_period_temporal
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return span_eq(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return span_eq(self._inner, spanset_span(other._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other) -> bool:
        """
        Returns whether ``self`` is strictly before ``other``. That is,
        ``self`` ends before ``other`` starts.

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            left_span_span, left_span_spanset
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return left_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return left_span_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_left(self, other) -> bool:
        """
        Returns whether ``self`` is before ``other`` allowing overlap. That is,
        ``self`` ends before ``other`` ends (or at the same time).

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overleft_span_span, overleft_span_spanset
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return overleft_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return overleft_span_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_right(self, other) -> bool:
        """
        Returns whether ``self`` is after ``other`` allowing overlap. That is,
        ``self`` starts after ``other`` starts (or at the same time).

        Args:
            other: temporal object to compare with

        Returns:
            True if overlapping or after, False otherwise

        MEOS Functions:
            overright_span_span, overright_span_spanset, overafter_period_timestamp,
            overafter_period_timestampset, overafter_period_temporal
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return overright_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return overright_span_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_right(self, other) -> bool:
        """
        Returns whether ``self`` is strictly after ``other``. That is, ``self``
        starts after ``other`` ends.

        Args:
            other: temporal object to compare with

        Returns:
            True if after, False otherwise

        MEOS Functions:
            right_span_span, right_span_spanset
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return right_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return right_span_spanset(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other) -> float:
        """
        Returns the distance between ``self`` and ``other``.

        Args:
            other:  object to compare with

        Returns:
            A :class:`flat` instance

        MEOS Functions:
            distance_span_span, distance_spanset_span
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return distance_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return distance_spanset_span(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Set Operations --------------------------------
    @abstractmethod
    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            A collection instance. The actual class depends on ``other``.

        MEOS Functions:
            intersection_span_span, intersection_spanset_span,
            intersection_period_timestamp
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return intersection_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return intersection_spanset_span(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def __mul__(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            A :class:`Span` instance. The actual class depends on ``other``.

        MEOS Functions:
            intersection_span_span, intersection_spanset_span,
            intersection_period_timestamp
        """
        return self.intersection(other)

    @abstractmethod
    def minus(self, other):
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: object to diff with

        Returns:
            A :class:`SpanSet` instance.
        """
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __sub__(self, other):
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: object to diff with

        Returns:
            A :class:`SpanSet` instance.
        """
        return self.minus(other)

    @abstractmethod
    def union(self, other):
        """
        Returns the temporal union of ``self`` and ``other``.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
        union_period_timestamp, union_spanset_span, union_span_span
        """
        from .spanset import SpanSet
        if isinstance(other, Span):
            return union_span_span(self._inner, other._inner)
        elif isinstance(other, SpanSet):
            return union_spanset_span(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def __add__(self, other):
        """
        Returns the temporal union of ``self`` and ``other``.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`PeriodSet` instance.

        MEOS Functions:
        union_period_timestamp, union_spanset_span, union_span_span
        """
        return self.union(other)

    # ------------------------- Comparisons -----------------------------------
    def __eq__(self, other):
        """
        Return whether ``self`` and ``other`` are equal.

        Args:
            other: temporal object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            span_eq
        """
        if isinstance(other, self.__class__):
            return span_eq(self._inner, other._inner)
        return False

    def __ne__(self, other):
        """
        Return whether ``self`` and ``other`` are not equal.

        Args:
            other: temporal object to compare with

        Returns:
            True if not equal, False otherwise

        MEOS Functions:
            span_neq
        """
        if isinstance(other, self.__class__):
            return span_ne(self._inner, other._inner)
        return True

    def __lt__(self, other):
        """
        Return whether ``self`` is less than ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if less than, False otherwise

        MEOS Functions:
            span_lt
        """
        if isinstance(other, self.__class__):
            return span_lt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __le__(self, other):
        """
        Return whether ``self`` is less than or equal to ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if less than or equal, False otherwise

        MEOS Functions:
            span_le
        """
        if isinstance(other, self.__class__):
            return span_le(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __gt__(self, other):
        """
        Return whether ``self`` is greater than ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if greater than, False otherwise

        MEOS Functions:
            span_gt
        """
        if isinstance(other, self.__class__):
            return span_gt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __ge__(self, other):
        """
        Return whether ``self`` is greater than or equal to ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            True if greater than or equal, False otherwise

        MEOS Functions:
            span_ge
        """
        if isinstance(other, self.__class__):
            return span_ge(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')



from .timestampset import TimestampSet
from .period import Period
from .periodset import PeriodSet
from .time import Time
from datetime import datetime, timedelta

__all__ = ['Time', 'TimestampSet', 'Period', 'PeriodSet', 'datetime', 'timedelta']



from __future__ import annotations

from datetime import datetime, timedelta
from typing import Optional, List, Union, TYPE_CHECKING, overload, get_args

from dateutil.parser import parse
from pymeos_cffi import *

from .time_collection import TimeCollection
from ..base import Set

if TYPE_CHECKING:
    from ...temporal import Temporal
    from .period import Period
    from .periodset import PeriodSet
    from .time import Time
    from ...boxes import Box


class TimestampSet(Set[datetime], TimeCollection):
    """
    Class for representing lists of distinct timestamp values.

    ``TimestampSet`` objects can be created with a single argument of type
    string as in MobilityDB.

        >>> TimestampSet(string='{2019-09-08 00:00:00+01, 2019-09-10 00:00:00+01, 2019-09-11 00:00:00+01}')

    Another possibility is to give a tuple or list of composing timestamps,
    which can be instances of ``str`` or ``datetime``. The composing timestamps
    must be given in increasing order.

        >>> TimestampSet(elements=['2019-09-08 00:00:00+01', '2019-09-10 00:00:00+01', '2019-09-11 00:00:00+01'])
        >>> TimestampSet(elements=[parse('2019-09-08 00:00:00+01'), parse('2019-09-10 00:00:00+01'), parse('2019-09-11 00:00:00+01')])

    """

    __slots__ = ['_inner']

    _mobilitydb_name = 'tstzset'

    _parse_function = timestampset_in
    _parse_value_function = lambda x: pg_timestamptz_in(x, -1) if isinstance(x, str) else datetime_to_timestamptz(x)
    _make_function = timestampset_make

    # ------------------------- Constructors ----------------------------------

    # ------------------------- Output ----------------------------------------
    def __str__(self):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            timestampset_out
        """
        return timestampset_out(self._inner)

    # ------------------------- Conversions -----------------------------------
    def to_spanset(self) -> PeriodSet:
        """
        Returns a PeriodSet that contains a Period for each Timestamp in
        ``self``.

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            set_to_spanset
        """
        from .periodset import PeriodSet
        return PeriodSet(_inner=super().to_spanset())

    def to_periodset(self) -> PeriodSet:
        """
        Returns a PeriodSet that contains a Period for each Timestamp in
        ``self``.

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            set_to_spanset
        """
        return self.to_spanset()

    def to_span(self) -> Period:
        """
        Returns a period that encompasses ``self``.

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            set_span
        """
        from .period import Period
        return Period(_inner=super().to_span())

    def to_period(self) -> Period:
        """
        Returns a period that encompasses ``self``.

        Returns:
            A new :class:`Period` instance

        MEOS Functions:
            set_span
        """
        return self.to_span()

    # ------------------------- Accessors -------------------------------------
    def duration(self) -> timedelta:
        """
        Returns the duration of the time ignoring gaps, i.e. the duration from
        the first timestamp to the last one.

        Returns:
            A :class:`datetime.timedelta` instance representing the duration of
            the period

        MEOS Functions:
            period_duration
        """
        return interval_to_timedelta(period_duration(set_span(self._inner)))

    def start_element(self) -> datetime:
        """
        Returns the first timestamp in ``self``.
        Returns:
            A :class:`datetime` instance

        MEOS Functions:
            timestampset_start_timestamp
        """
        return timestamptz_to_datetime(timestampset_start_timestamp(self._inner))

    def end_element(self) -> datetime:
        """
        Returns the last timestamp in ``self``.
        Returns:
            A :class:`datetime` instance

        MEOS Functions:
            timestampset_end_timestamp
        """
        return timestamptz_to_datetime(timestampset_end_timestamp(self._inner))

    def element_n(self, n: int) -> datetime:
        """
        Returns the n-th timestamp in ``self``.
        Returns:
            A :class:`datetime` instance

        MEOS Functions:
            timestampset_timestamp_n
        """
        super().element_n(n)
        return timestamptz_to_datetime(timestampset_timestamp_n(self._inner, n + 1))

    def elements(self) -> List[datetime]:
        """
        Returns the list of distinct timestamps in ``self``.
        Returns:
            A :class:`list[datetime]` instance

        MEOS Functions:
            timestampset_timestamps
        """
        tss = timestampset_values(self._inner)
        return [timestamptz_to_datetime(tss[i]) for i in range(self.num_elements())]

    # ------------------------- Transformations -------------------------------
    def shift(self, delta: timedelta) -> TimestampSet:
        """
        Returns a new TimestampSet that is the result of shifting ``self`` by
        ``delta``

        Examples:
            >>> TimestampSet('{2000-01-01, 2000-01-10}').shift(timedelta(days=2))
            >>> 'TimestampSet({2000-01-03 00:00:00+01, 2000-01-12 00:00:00+01})'

        Args:
            delta: :class:`datetime.timedelta` instance to shift

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            timestampset_shift_scale
        """
        return self.shift_scale(shift=delta)

    def scale(self, duration: timedelta) -> TimestampSet:
        """
        Returns a new TimestampSet that with the scaled so that the span of
        ``self`` is ``duration``.

        Examples:
            >>> TimestampSet('{2000-01-01, 2000-01-10}').scale(timedelta(days=2))
            >>> 'TimestampSet({2000-01-01 00:00:00+01, 2000-01-03 00:00:00+01})'

        Args:
            duration: :class:`datetime.timedelta` instance representing the
            span of the new set

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            timestampset_shift_scale
        """
        return self.shift_scale(duration=duration)

    def shift_scale(self, shift: Optional[timedelta] = None,
        duration: Optional[timedelta] = None) -> TimestampSet:
        """
        Returns a new TimestampSet that is the result of shifting and scaling
        ``self``.

        Examples:
            >>> TimestampSet('{2000-01-01, 2000-01-10}').shift_scale(shift=timedelta(days=2), duration=timedelta(days=4))
            >>> 'TimestampSet({2000-01-03 00:00:00+01, 2000-01-07 00:00:00+01})'

        Args:
            shift: :class:`datetime.timedelta` instance to shift
            duration: :class:`datetime.timedelta` instance representing the
            span of the new set

        Returns:
            A new :class:`PeriodSet` instance

        MEOS Functions:
            timestampset_shift_scale
        """
        assert shift is not None or duration is not None, 'shift and scale deltas must not be both None'
        tss = timestampset_shift_scale(
            self._inner,
            timedelta_to_interval(shift) if shift else None,
            timedelta_to_interval(duration) if duration else None
        )
        return TimestampSet(_inner=tss)

    # ------------------------- Topological Operations ------------------------
    def is_adjacent(self, other: Union[Period, PeriodSet, Temporal, Box]) -> bool:
        """
        Returns whether ``self`` is temporally adjacent to ``other``. That is,
        they share a bound but only one of them contains it.

        Examples:
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_adjacent(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_adjacent(Period('[2012-01-02, 2012-01-03]'))
            >>> False  # Both contain bound
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_adjacent(Period('(2012-01-02, 2012-01-03]'))
            >>> False  # Neither contain bound

        Args:
            other: temporal object to compare with

        Returns:
            True if adjacent, False otherwise

        MEOS Functions:
            adjacent_span_span, adjacent_spanset_span
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, Temporal):
            return self.is_adjacent(other.time())
        elif isinstance(other, get_args(Box)):
            return self.is_adjacent(other.to_period())
        else:
            super().is_adjacent(other)

    def is_contained_in(self, container: Union[Period, PeriodSet, TimestampSet,
        Temporal, Box]) -> bool:
        """
        Returns whether ``self`` is temporally contained in ``container``.

        Examples:
            >>> TimestampSet('{2012-01-02, 2012-01-03}').is_contained_in(Period('[2012-01-01, 2012-01-04]'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_contained_in(Period('[2012-01-01, 2012-01-02]'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_contained_in(Period('(2012-01-01, 2012-01-02)'))
            >>> False

        Args:
            container: temporal object to compare with

        Returns:
            True if contained, False otherwise

        MEOS Functions:
            contained_span_span, contained_span_spanset, contained_set_set,
            contained_spanset_spanset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(container, Temporal):
            return self.is_contained_in(container.time())
        elif isinstance(container, Box):
            return self.is_contained_in(container.to_period())
        else:
            return super().is_contained_in(container)

    def contains(self, content: Union[datetime, TimestampSet, Temporal]) -> bool:
        """
        Returns whether ``self`` temporally contains ``content``.

        Examples:
            >>> TimestampSet('{2012-01-01, 2012-01-04}').contains(parse('2012-01-01]'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').contains(TimestampSet('{2012-01-01}'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').contains(TimestampSet('{2012-01-01, 2012-01-03}'))
            >>> False

        Args:
            content: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_timestampset_timestamp, contains_set_set, 
            contains_spanset_spanset
        """
        from ...temporal import Temporal
        if isinstance(content, datetime):
            return contains_timestampset_timestamp(self._inner,
                datetime_to_timestamptz(content))
        elif isinstance(content, Temporal):
            return self.to_spanset().contains(content)
        else:
            return super().contains(content)

    def __contains__(self, item):
        """
        Returns whether ``self`` temporally contains ``content``.

        Examples:
            >>> TimestampSet('{2012-01-01, 2012-01-04}').contains(parse('2012-01-01]'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').contains(TimestampSet('{2012-01-01}'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').contains(TimestampSet('{2012-01-01, 2012-01-03}'))
            >>> False

        Args:
            item: temporal object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_timestampset_timestamp, contains_set_set, 
            contains_spanset_spanset
        """
        return self.contains(item)

    def overlaps(self, other: Union[Period, PeriodSet, TimestampSet, Temporal,
        Box]) -> bool:
        """
        Returns whether ``self`` temporally overlaps ``other``. That is, both
        share at least an instant

        Examples:
            >>> TimestampSet('{2012-01-01, 2012-01-02}').overlaps(TimestampSet('{2012-01-02, 2012-01-03}'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').overlaps(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').overlaps(Period('(2012-01-02, 2012-01-03]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if overlaps, False otherwise

        MEOS Functions:
            overlaps_set_set, overlaps_span_span, overlaps_spanset_spanset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return contains_timestampset_timestamp(self._inner,
                datetime_to_timestamptz(other))
        elif isinstance(other, Temporal):
            return self.to_spanset().overlaps(other)
        elif isinstance(other, Box):
            return self.to_span().overlaps(other)
        else:
            return super().overlaps(other)

    def is_same(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether the bounding period of `self` is the same as the
        bounding period of `other`.

        Args:
            other: A time or temporal object to compare to `self`.

        Returns:
            True if same, False otherwise.

        See Also:
            :meth:`Period.is_same`
        """
        return self.to_period().is_same(other)

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether ``self`` is strictly before ``other``. That is,
        ``self`` ends before ``other`` starts.

        Examples:
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_left(TimestampSet('{2012-01-03}'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_left(Period('(2012-01-02, 2012-01-03]'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_left(Period('[2012-01-02, 2012-01-03]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overafter_timestamp_period, left_span_span, left_span_spanset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return after_timestamp_timestampset(datetime_to_timestamptz(other),
                self._inner)
        elif isinstance(other, Temporal):
            return self.to_period().is_left(other)
        elif isinstance(other, get_args(Box)):
            return self.to_period().is_left(other)
        else:
            return super().is_left(other)

    def is_over_or_left(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether ``self`` is before ``other`` allowing overlap. That is,
        ``self`` ends before ``other`` ends (or at the same time).

        Examples:
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_over_or_left(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> TimestampSet('{2012-01-01, 2012-01-02}').is_over_or_left(Period('[2012-01-02, 2012-01-03]'))
            >>> True
            >>> TimestampSet('{2012-01-03, 2012-01-05}').is_over_or_left(Period('[2012-01-01, 2012-01-04]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overbefore_period_timestamp, overleft_span_span, overleft_span_spanset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overafter_timestamp_timestampset(datetime_to_timestamptz(other),
                self._inner)
        elif isinstance(other, Temporal):
            return self.to_period().is_over_or_left(other)
        elif isinstance(other, get_args(Box)):
            return self.to_period().is_over_or_left(other.to_period())
        else:
            return super().is_over_or_left(other)

    def is_over_or_right(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether ``self`` is after ``other`` allowing overlap. That is,
        ``self`` starts after ``other`` starts (or at the same time).

        Examples:
            >>> TimestampSet('{2012-01-02, 2012-01-03}').is_over_or_right(Period('[2012-01-01, 2012-01-02)'))
            >>> True
            >>> TimestampSet('{2012-01-02, 2012-01-03}').is_over_or_right(Period('[2012-01-01, 2012-01-02]'))
            >>> True
            >>> TimestampSet('{2012-01-02, 2012-01-03}').is_over_or_right(Period('[2012-01-01, 2012-01-03]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if overlapping or after, False otherwise

        MEOS Functions:
        overafter_period_timestamp, overright_span_span, overright_span_spanset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return overbefore_timestamp_timestampset(datetime_to_timestamptz(other),
                self._inner)
        elif isinstance(other, Temporal):
            return self.to_period().is_over_or_right(other)
        elif isinstance(other, get_args(Box)):
            return self.to_period().is_over_or_right(other)
        else:
            return super().is_over_or_right(other)

    def is_right(self, other: Union[Time, Temporal, Box]) -> bool:
        """
        Returns whether ``self`` is strictly after ``other``. That is, the
        first timestamp in ``self`` is after ``other``.

        Examples:
            >>> TimestampSet('{2012-01-02, 2012-01-03}').is_right(Period('[2012-01-01, 2012-01-02)'))
            >>> True
            >>> TimestampSet('{2012-01-02, 2012-01-03}').is_right(TimestampSet('{2012-01-01}'))
            >>> True
            >>> TimestampSet('{2012-01-02, 2012-01-03}').is_right(Period('[2012-01-01, 2012-01-02]'))
            >>> False

        Args:
            other: temporal object to compare with

        Returns:
            True if after, False otherwise

        MEOS Functions:
            overbefore_timestamp_timestampset, right_set_set, right_span_span, 
            right_span_spanset
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return before_timestamp_timestampset(datetime_to_timestamptz(other),
                self._inner)
        elif isinstance(other, Temporal):
            return self.to_period().is_right(other)
        elif isinstance(other, get_args(Box)):
            return self.to_period().is_right(other)
        else:
            return super().is_right(other)

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other: Union[Time, Temporal, Box]) -> timedelta:
        """
        Returns the temporal distance between ``self`` and ``other``.

        Args:
            other: temporal object to compare with

        Returns:
            A :class:`datetime.timedelta` instance

        MEOS Functions:
            distance_timestampset_timestamp, distance_set_set, 
            distance_span_span, distance_spanset_span
        """
        from ...temporal import Temporal
        from ...boxes import Box
        if isinstance(other, datetime):
            return timedelta(seconds=distance_timestampset_timestamp(self._inner, datetime_to_timestamptz(other)))
        elif isinstance(other, Temporal):
            return self.to_period().distance(other)
        elif isinstance(other, get_args(Box)):
            return self.to_period().distance(other)
        else:
            return timedelta(seconds=super().distance(other))

    # ------------------------- Set Operations --------------------------------
    @overload
    def intersection(self, other: datetime) -> Optional[datetime]:
        ...

    @overload
    def intersection(self, other: TimestampSet) -> Optional[TimestampSet]:
        ...

    @overload
    def intersection(self, other: Union[Period, PeriodSet, Temporal, Box]) -> Optional[PeriodSet]:
        ...

    def intersection(self, other: Union[Time, Temporal]) -> Optional[Time]:
        """
        Returns the temporal intersection of ``self`` and ``other``.

        Args:
            other: temporal object to intersect with

        Returns:
            A :class:`Time` instance. The actual class depends on ``other``.

        MEOS Functions:
            intersection_set_set, intersection_spanset_span, 
            intersection_spanset_spanset
        """
        from .period import Period
        from .periodset import PeriodSet
        if isinstance(other, datetime):
            result = intersection_timestampset_timestamp(self._inner,
                datetime_to_timestamptz(other))
            return timestamptz_to_datetime(result) if result is not None else None
        elif isinstance(other, TimestampSet):
            result = intersection_set_set(self._inner, other._inner)
            return TimestampSet(_inner=result) if result is not None else None
        elif isinstance(other, Period):
            return self.to_periodset().intersection(other)
        elif isinstance(other, PeriodSet):
            return self.to_periodset().intersection(other)
        elif isinstance(other, Temporal):
            return self.intersection(other.time())
        elif isinstance(other, get_args(Box)):
            return self.intersection(other.to_period())
        else:
            return super().intersection(other)

    @overload
    def minus(self, other: Union[datetime, TimestampSet]) -> Optional[TimestampSet]:
        ...

    @overload
    def minus(self, other: Union[Period, PeriodSet, Temporal, Box]) -> Optional[PeriodSet]:
        ...

    def minus(self, other: Union[Time, Temporal, Box]) -> Optional[Time]:
        """
        Returns the temporal difference of ``self`` and ``other``.

        Args:
            other: temporal object to diff with

        Returns:
            A :class:`Time` instance. The actual class depends on ``other``.

        MEOS Functions:
            minus_timestampset_timestamp, minus_set_set, minus_spanset_span, 
            minus_spanset_spanset
        """
        from .period import Period
        from .periodset import PeriodSet
        if isinstance(other, datetime):
            result = minus_timestampset_timestamp(self._inner,
                datetime_to_timestamptz(other))
            return TimestampSet(_inner=result) if result is not None else None
        elif isinstance(other, TimestampSet):
            result = minus_set_set(self._inner, other._inner)
            return TimestampSet(_inner=result) if result is not None else None
        elif isinstance(other, Period):
            return self.to_periodset().minus(other)
        elif isinstance(other, PeriodSet):
            return self.to_periodset().minus(other)
        elif isinstance(other, Temporal):
            return self.minus(other.time())
        elif isinstance(other, get_args(Box)):
            return self.minus(other.to_period())
        else:
            return super().minus(other)

    def subtract_from(self, other: datetime) -> Optional[datetime]:
        """
        Returns the difference of ``other`` and ``self``.

        Args:
            other: A :class:`datetime` instance

        Returns:
            A :class:`datetime` instance or ``None`` if the difference is empty.

        MEOS Functions:
            minus_timestamp_timestampset

        See Also:
            :meth:`minus`
        """
        return timestamptz_to_datetime(minus_timestamp_timestampset(datetime_to_timestamptz(other), self._inner))

    @overload
    def union(self, other: Union[datetime, TimestampSet]) -> TimestampSet:
        ...

    @overload
    def union(self, other: Union[Period, PeriodSet, Temporal, Box]) -> PeriodSet:
        ...

    def union(self, other: Union[Time, Temporal, Box]) -> Union[PeriodSet, TimestampSet]:
        """
        Returns the temporal union of ``self`` and ``other``.

        Args:
            other: temporal object to merge with

        Returns:
            A :class:`Time` instance. The actual class depends on ``other``.

        MEOS Functions:
            union_timestampset_timestamp, union_set_set, union_spanset_span, 
            union_spanset_spanset
        """
        from .period import Period
        from .periodset import PeriodSet
        if isinstance(other, datetime):
            return TimestampSet(_inner=union_timestampset_timestamp(self._inner, datetime_to_timestamptz(other)))
        elif isinstance(other, TimestampSet):
            return TimestampSet(_inner=union_set_set(self._inner, other._inner))
        elif isinstance(other, Period):
            return self.to_periodset().union(other)
        elif isinstance(other, PeriodSet):
            return self.to_periodset().union(other)
        elif isinstance(other, Temporal):
            return self.union(other.time())
        elif isinstance(other, get_args(Box)):
            return self.union(other.to_period())
        else:
            return super().union(other)

    # ------------------------- Comparisons -----------------------------------

    # ------------------------- Plot Operations -------------------------------
    def plot(self, *args, **kwargs):
        from ...plotters import TimePlotter
        return TimePlotter.plot_timestampset(self, *args, **kwargs)



from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..main import TText


class TemporalTextMaxAggregator(BaseAggregator[TText, TText]):
    """
    Temporal maximum of all aggregated temporal texts.

    MEOS Functions:
        ttext_tmax_transfn, temporal_tagg_finalfn
    """
    _add_function = ttext_tmax_transfn


class TemporalTextMinAggregator(BaseAggregator[TText, TText]):
    """
    Temporal minimum of all aggregated temporal texts.

    MEOS Functions:
        ttext_tmin_transfn, temporal_tagg_finalfn
    """
    _add_function = ttext_tmin_transfn



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Union, List
from typing import TypeVar, Type, Callable, Any, TYPE_CHECKING, Iterable

from pymeos_cffi import *

from .collection import Collection

if TYPE_CHECKING:
    from .spanset import SpanSet
    from .span import Span

T = TypeVar('T')
Self = TypeVar('Self', bound='Set[Any]')


class Set(Collection[T], ABC):
    """
    Base class for all set classes.
    """

    __slots__ = ['_inner']

    _parse_function: Callable[[str], 'CData'] = None
    _parse_value_function: Callable[[Union[str, T]], Any] = None
    _make_function: Callable[[Iterable[Any]], 'CData'] = None

    # ------------------------- Constructors ----------------------------------
    def __init__(self, string: Optional[str] = None, *,
                 elements: Optional[List[Union[str, T]]] = None,
                 _inner=None):
        super().__init__()
        assert (_inner is not None) or ((string is not None) != (elements is not None)), \
            "Either string must be not None or elements must be not"
        if _inner is not None:
            self._inner = _inner
        elif string is not None:
            self._inner = self.__class__._parse_function(string)
        else:
            parsed_elements = [self.__class__._parse_value_function(ts) for ts in elements]
            self._inner = self.__class__._make_function(parsed_elements)

    def __copy__(self: Self) -> Self:
        """
        Return a copy of ``self``.

        Returns:
            A new :class:`Span` instance

        MEOS Functions:
            set_copy
        """
        inner_copy = set_copy(self._inner)
        return self.__class__(_inner=inner_copy)

    @classmethod
    def from_wkb(cls: Type[Self], wkb: bytes) -> Self:
        """
        Returns a `Set` from its WKB representation.
        Args:
            wkb: WKB representation

        Returns:
            A new :class:`Set` instance

        MEOS Functions:
            set_from_wkb
        """
        return cls(_inner=set_from_wkb(wkb))

    @classmethod
    def from_hexwkb(cls: Type[Self], hexwkb: str) -> Self:
        """
        Returns a `Set` from its WKB representation in hex-encoded ASCII.
        Args:
            hexwkb: WKB representation in hex-encoded ASCII

        Returns:
            A new :class:`Set` instance

        MEOS Functions:
            set_from_hexwkb
        """
        return cls(_inner=(set_from_hexwkb(hexwkb)))

    # ------------------------- Output ----------------------------------------
    @abstractmethod
    def __str__(self):
        """
        Return the string representation of the content of ``self``.

        Returns:
            A new :class:`str` instance
        """
        raise NotImplementedError()

    def __repr__(self):
        """
        Return the string representation of ``self``.

        Returns:
            A new :class:`str` instance

        MEOS Functions:
            set_out
        """
        return (f'{self.__class__.__name__}'
                f'({self})')

    def as_wkb(self) -> bytes:
        """
        Returns the WKB representation of ``self``.
        Returns:
            A :class:`str` object with the WKB representation of ``self``.

        MEOS Functions:
            set_as_wkb
        """
        return set_as_wkb(self._inner, 4)

    def as_hexwkb(self) -> str:
        """
        Returns the WKB representation of ``self`` in hex-encoded ASCII.
        Returns:
            A :class:`str` object with the WKB representation of ``self`` in hex-encoded ASCII.

        MEOS Functions:
            set_as_hexwkb
        """
        return set_as_hexwkb(self._inner, -1)[0]

    # ------------------------- Conversions -----------------------------------
    @abstractmethod
    def to_spanset(self) -> SpanSet:
        """
        Returns a SpanSet that contains a Span for each element in ``self``.

        Returns:
            A new :class:`SpanSet` instance

        MEOS Functions:
            set_to_spanset
        """
        return set_to_spanset(self._inner)

    @abstractmethod
    def to_span(self) -> Span:
        """
        Returns a span that encompasses ``self``.

        Returns:
            A new :class:`Span` instance

        MEOS Functions:
            set_span
        """
        return set_span(self._inner)

    # ------------------------- Accessors -------------------------------------

    def num_elements(self) -> int:
        """
        Returns the number of elements in ``self``.
        Returns:
            An :class:`int`

        MEOS Functions:
            set_num_values
        """
        return set_num_values(self._inner)

    def __len__(self):
        """
        Returns the number of elements in ``self``.
        Returns:
            An :class:`int`

        MEOS Functions:
            set_num_values
        """
        return self.num_elements()

    @abstractmethod
    def start_element(self) -> T:
        """
        Returns the first element in ``self``.
        Returns:
            A :class:`T` instance
        """
        raise NotImplementedError()

    @abstractmethod
    def end_element(self) -> T:
        """
        Returns the last element in ``self``.
        Returns:
            A :class:`T` instance
        """
        raise NotImplementedError()

    @abstractmethod
    def element_n(self, n: int) -> T:
        """
        Returns the n-th element in ``self``.
        Returns:
            A :class:`T` instance
        """
        if n < 0 or n >= self.num_elements():
            raise IndexError(f'Index {n} out of bounds')

    @abstractmethod
    def elements(self) -> List[T]:
        """
        Returns the list of distinct elements in ``self``.
        Returns:
            A :class:`list[T]` instance
        """
        raise NotImplementedError()

    def __hash__(self) -> int:
        """
        Return the hash representation of ``self``.

        Returns:
            A new :class:`int` instance

        MEOS Functions:
            set_hash
        """
        return set_hash(self._inner)

    # ------------------------- Topological Operations ------------------------

    def is_contained_in(self, container) -> bool:
        """
        Returns whether ``self`` is contained in ``container``.

        Args:
            container: object to compare with

        Returns:
            True if contained, False otherwise

        MEOS Functions:
            contained_span_span, contained_span_spanset, contained_set_set,
            contained_spanset_spanset
        """
        if isinstance(container, Set):
            return contained_set_set(self._inner, container._inner)
        else:
            raise TypeError(f'Operation not supported with type {container.__class__}')

    @abstractmethod
    def contains(self, content) -> bool:
        """
        Returns whether ``self`` contains ``content``.

        Args:
            content: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_set_set
        """
        if isinstance(content, Set):
            return contains_set_set(self._inner, content._inner)
        else:
            raise TypeError(f'Operation not supported with type {content.__class__}')

    def __contains__(self, item):
        """
        Returns whether ``self`` contains ``content``.

        Args:
            item: object to compare with

        Returns:
            True if contains, False otherwise

        MEOS Functions:
            contains_set_set
        """
        return self.contains(item)

    def overlaps(self, other) -> bool:
        """
        Returns whether ``self`` overlaps ``other``. That is, both share at least an instant

        Args:
            other: object to compare with

        Returns:
            True if overlaps, False otherwise

        MEOS Functions:
            overlaps_set_set, overlaps_span_span, overlaps_spanset_spanset
        """
        if isinstance(other, Set):
            return overlaps_set_set(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Position Operations ---------------------------
    def is_left(self, other) -> bool:
        """
        Returns whether ``self`` is strictly to the left of ``other``. That is,
        ``self`` ends before ``other`` starts.

        Args:
            other: object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            left_span_span, left_span_spanset
        """
        if isinstance(other, Set):
            return left_set_set(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_left(self, other) -> bool:
        """
        Returns whether ``self`` is to the left of ``other`` allowing overlap.
        That is, ``self`` ends before ``other`` ends (or at the same value).

        Args:
            other: object to compare with

        Returns:
            True if before, False otherwise

        MEOS Functions:
            overleft_span_span, overleft_span_spanset
        """
        if isinstance(other, Set):
            return overleft_set_set(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_right(self, other) -> bool:
        """
        Returns whether ``self`` is to the right of ``other`` allowing overlap. That is, ``self`` starts after ``other``
        starts (or at the same value).

        Args:
            other: object to compare with

        Returns:
            True if overlapping or to the right, False otherwise

        MEOS Functions:
            overright_span_span, overright_span_spanset
        """
        if isinstance(other, Set):
            return overright_set_set(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_right(self, other) -> bool:
        """
        Returns whether ``self`` is strictly to the right of ``other``. That is, the first element in ``self``
        is to the right ``other``.

        Args:
            other: object to compare with

        Returns:
            True if right, False otherwise

        MEOS Functions:
            right_set_set, right_span_span, right_span_spanset
        """
        if isinstance(other, Set):
            return right_set_set(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Distance Operations ---------------------------
    def distance(self, other) -> float:
        """
        Returns the distance between ``self`` and ``other``.

        Args:
            other: object to compare with

        Returns:
            A :class:`float` instance

        MEOS Functions:
            distance_set_set, distance_span_span, distance_spanset_span
        """
        from .span import Span
        from .spanset import SpanSet
        if isinstance(other, Set):
            return distance_set_set(self._inner, other._inner)
        elif isinstance(other, Span):
            return distance_span_span(set_span(self._inner), other._inner)
        elif isinstance(other, SpanSet):
            return distance_spanset_span(other._inner, set_span(self._inner))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    # ------------------------- Set Operations --------------------------------
    @abstractmethod
    def intersection(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            A :class:`Collection` instance. The actual class depends on ``other``.

        MEOS Functions:
            intersection_set_set, intersection_spanset_span,
            intersection_spanset_spanset
        """
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __mul__(self, other):
        """
        Returns the intersection of ``self`` and ``other``.

        Args:
            other: object to intersect with

        Returns:
            A :class:`Collection` instance. The actual class depends on ``other``.
        """
        return self.intersection(other)

    @abstractmethod
    def minus(self, other):
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: object to diff with

        Returns:
            A :class:`Collection` instance. The actual class depends on ``other``.
        """
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __sub__(self, other):
        """
        Returns the difference of ``self`` and ``other``.

        Args:
            other: object to diff with

        Returns:
            A :class:`Collection` instance. The actual class depends on ``other``.
        """
        return self.minus(other)

    @abstractmethod
    def subtract_from(self, other):
        """
        Returns the difference of ``other`` and ``self``.

        Args:
            other: object to subtract ``self`` from

        Returns:
            A :class:`Collection` instance or an element instance. The actual class depends on ``other``.

        See Also:
            :meth:`minus`
        """
        raise NotImplementedError()

    def __rsub__(self, other):
        """
        Returns the difference of ``other`` and ``self``.

        Args:
            other: object to subtract ``self`` from

        Returns:
            A :class:`Collection` instance or an element instance. The actual class depends on ``other``.
        """
        return self.subtract_from(other)

    @abstractmethod
    def union(self, other):
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: object to merge with

        Returns:
            A :class:`Collection` instance. The actual class depends on ``other``.
        """
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __add__(self, other):
        """
        Returns the union of ``self`` and ``other``.

        Args:
            other: object to merge with

        Returns:
            A :class:`Collection` instance. The actual class depends on ``other``.
        """
        return self.union(other)

    # ------------------------- Comparisons -----------------------------------
    def __eq__(self, other):
        """
        Returns whether ``self`` and ``other`` are equal.

        Args:
            other: object to compare with

        Returns:
            True if equal, False otherwise

        MEOS Functions:
            set_eq
        """
        if isinstance(other, self.__class__):
            return set_eq(self._inner, other._inner)
        return False

    def __ne__(self, other):
        """
        Returns whether ``self`` and ``other`` are not equal.

        Args:
            other: object to compare with

        Returns:
            True if not equal, False otherwise

        MEOS Functions:
            set_ne
        """
        if isinstance(other, self.__class__):
            return set_ne(self._inner, other._inner)
        return True

    def __lt__(self, other):
        """
        Return whether ``self`` is less than ``other``.

        Args:
            other: object to compare with

        Returns:
            True if less than, False otherwise

        MEOS Functions:
            set_lt
        """
        if isinstance(other, self.__class__):
            return set_lt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __le__(self, other):
        """
        Return whether ``self`` is less than or equal to ``other``.

        Args:
            other: object to compare with

        Returns:
            True if less than or equal, False otherwise

        MEOS Functions:
            set_le
        """
        if isinstance(other, self.__class__):
            return set_le(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __gt__(self, other):
        """
        Return whether ``self`` is greater than ``other``.

        Args:
            other: object to compare with

        Returns:
            True if greater than, False otherwise

        MEOS Functions:
            set_gt
        """
        if isinstance(other, self.__class__):
            return set_gt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __ge__(self, other):
        """
        Return whether ``self`` is greater than or equal to ``other``.

        Args:
            other: object to compare with

        Returns:
            True if greater than or equal, False otherwise

        MEOS Functions:
            set_ge
        """
        if isinstance(other, self.__class__):
            return set_ge(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')
